'From Squeak3.4gamma of ''7 January 2003'' [latest update: #5168] on 22 February 2003 at 10:29:21 pm'!
Error subclass: #ArithmeticError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions Kernel'!
Magnitude subclass: #DateAndTime
	instanceVariableNames: 'localDate localOffsetToUTC localTime localTimeZoneName localTimeZoneAbbrev '
	classVariableNames: 'DaylightEndDate DaylightOffset DaylightStartDate DaylightTimeZoneAbbrev DaylightTimeZoneName DaysInMonth DaysToMonth DefaultIsStandard Offset StandardOffset StandardTimeZoneAbbrev StandardTimeZoneName TimeZoneAbbrev TimeZoneName '
	poolDictionaries: ''
	category: 'Kernel-Magnitudes'!

!DateAndTime commentStamp: '<historical>' prior: 0!
DateAndTime instances represent individual points in Coordinated Universal Time (UTC) formerly called GMT at Greenwich, UK as represented in an implementation defined local time.  
	DateAndTime is an ANSI standard class intended to replace the Blue Book Date and Time classes, providing proper facilities for DateAndTime representation and manipulation, including consideration of time zones.

Typical Use:

	DateAndTime instances are suitable for representing file creation date and time, or revision change control (RCS) change date and time.

Compliance:

	DateAndTime complies with the following ANSI protocols:

	Object
	magnitude
	DateAndTime

Open Issues:

	#asUTC - What time zone name is the time returned.

class

	#clockPrecision - What should be returned.  It most likely
			depends on the operating system.
	#initialize - It would be useful to initialization the default offset,
			time zone name and abbreviation variables from the
			operating system or environent settings.

Implementation:

	DateAndTime relies heavily on the existing Date & Duration class functions for calculating various values such as day of the week, etc.
	Each <DateAndTime> has an associated offset to UTC of his local time, the time zone name and abbreviation.  These are set from standard time or daylight saving time defaults determined by DefaultIsStandard switch.
	These defaults are set at initialization by the user to those appropriate for his local time.
	RuleDate class instances representing the first & last day of daylight saving time calculate the exact dates for a given year.  Each time a default offset is requested the first & last day of daylight saving time exact year is checked to see if it is the same as today's year.  If not new exact dates for today's year are calculated and the RuleDate instances are updated.
	During this offset request it also checks when today's date is during the daylight saving time period.  If standard time then the default standard time values are returned, else if daylight saving time then the default daylight saving time values.

Instance variables:

	localDate	- a Date class instance representing the specified date
				in local time.
	localTime	- a <Duration> representing the specified time
				in local time.
	localOffsetToUTC	- a <Duration> representing an amount of time
				(-west/+east) local time differs from UTC.
	localTimeZoneName	- a <readableString> representing the name of
				the time zone in local time.
	localTimeZoneAbbrev	- a <readableString> representing the
				abbreviation of the name of the time zone in local time.

Class variables:

	StandardOffset	- a <Duration> representing the default amount
				of time (-west/+east) this users local standard time
				differs from UTC.
	StandardTimeZoneName	- a <readableString> representing the default
				name of the time zone of local standard time.
	StandardTimeZoneAbbrev	- a <readableString> representing the
				default abbreviation of the name of the time zone in
				local standard time.
	DaylightOffset	- a <Duration> representing the default amount
				of time (-west/+east) this users local daylight saving
				time differs from UTC.
	DaylightTimeZoneName	- a <readableString> representing the default
				name of the time zone of local daylight saving time.
	DaylightTimeZoneAbbrev	- a <readableString> representing the
				default abbreviation of the name of the time zone in
				local daylight saving time.
	DaylightStartDate	- a RuleDate class instance representing the
				first day of daylight saving time.
	DaylightEndDate	- a RuleDate class instance representing the last 
				day of daylight saving time.
	DefaultIsStandard	- a <boolean> indicating if true that today's date
				(when last checked) is during local standard time,
				if false, daylight saving time.  This is normally used
				to determine if standard time or daylight saving time
				defaults should be returned.


Assumptions:

	It's assumed that an instance is an indivisible whole.  If the date, time, or offset is changed, a new instance is created with the new aspect.
	It's assumed that a new instance is rationalized.  This means the instance is adjusted, if necessary, so localTime is less than a day and positive.
	It's assumed that the offset is zero at Greenwich, and decreases one hour for each time zone the locale is west of Greenwich and increases to the east.  As an example, if the locale is New York, NY, in the Eastern Standard Time (EST) it would have an offset of 5 hours.

				Standard Time
	<- West	PST  MST CST  EST  ???  ???   ???  ???  UTC   CET	East ->
			---   ---  ---   ---  ---   ---   ---  ---  ----   ----
	time	8PM 7PM 6PM 5PM 4PM 3PM 2PM 1PM noon 11am
	offset	-8   -7    -6   -5    -4   -3   -2    -1   0      +1

Conversion between UTC and local time:

	Standard Time:

		UTC			= local time - offset 
		local time	= UTC + offset

		Central Standard Time - USA = UTC - 6

	Daylight Saving Time

		local time	= UTC + (Standard offset + 1)

		Central Daylight Time - USA = UTC - 5

	Most helpful web site on conversion:

		http://ww2010.atmos.uiuc.edu/(Gl)/guides/maps/utc/frutc.rxml

	Most helpful web site that includes a brief overview of ISO 8601 on which DateAndTime is based:

		http://www.cl.cam.ac.uk/~mgk25/iso-time.html

		A Summary of the International Standard Date and Time Notation
			by Markus Kuhn

Locale Initialization:

	Local time zone abbreviation.
	Local time zone name.
	Offset (-west/+east) between the local time and UTC.

Additions:


Alternate Implementation:

	I nearly finished an implementation that stored the date in UTC, instead of local time, but abandoned this approach because I seemed to create a temporary DateAndTime instance for almost every ANSI message.  I will keep this UTC implementation and will gladly send it to anyone that want. it.
!

Magnitude subclass: #Duration
	instanceVariableNames: 'asSeconds '
	classVariableNames: 'SecondsInDay SecondsInHour SecondsInMinute '
	poolDictionaries: ''
	category: 'Kernel-Magnitudes'!

!Duration commentStamp: '<historical>' prior: 0!
DateAndTime instances represent periods of time.

Duration is an ANSI standard class intended, in conjunction with DateAndTime, to replace the Blue Book Date and Time classes, providing proper facilities for DateAndTime representation and manipulation, including consideration of time zones.

Typical Use:

	Duration instances are suitable for representing a length of time such as running a test.

Duration complies with the following ANSI protocols:

	Object
	magnitude
	Duration

The Duration class itself supports the <Duration factory> protocol.


Implementation:

Instance variables:

	asSeconds	- <Float> | <Integer> representing the duration length
				in seconds.

Class variables:

	SecondsInDay	- <Integer> representing the seconds in one day.
	SecondsInHour	- <Integer> representing the seconds in one hour.
	SecondsInMinute	- <Integer> representing the seconds in one minute.
!

ArithmeticError subclass: #FloatingPointException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions Kernel'!

!Number commentStamp: '<historical>' prior: 0!
I am an abstract representation of a number.  My subclasses Float, Fraction, Integer, and ScaledDecimal, and their subclasses, provide concrete representations of a numeric quantity.
	All my subclasses participate in a simple type coercion mechanism that supports mixed-mode arithmetic and comparisons.  It works as follows:  If

	self<typeA> op: arg<typeB>

fails because of incompatible types, then it is retried in the following guise:

	^arg
		adaptToTypeA: self
		andSend: #'op:'

	This gives the arg of typeB an opportunity to resolve the incompatibility, knowing exactly what two types are involved.  This mechanism is extensible to any new number classes that one might wish to add to Squeak.  The only requirement is that every subclass of Number must support the conversion method specific to each of the other subclasses of Number.
	Unless otherwise specified by an individual operation the receiver and argument are first converted to the same numeric representation according to the following table similar to the table in section 5.6.2 Protocol: <number> of the Draft American National Standard for Information Systems - Programming Languages - Smalltalk:

						Default Conversion Table:
				(simplified for Squeak's single Float type) 
						
		operand |	<integer>	<scaledDec>	<Fraction>	<Float>(e|d|q)
	--receiver---|-------------------------------------------------------
	<integer>	|	<integer>	<scaledDec>	<Fraction>	<Float>e
	<scaledDec>	|	<scaledDec>	<scaledDec>	<Fraction>	<Float>e
	<Fraction>	|	<Fraction>	<Fraction>	<Fraction>	<Float>e
	<Float>(e|d|q)	|	<Float>e		<Float>e		<Float>e		<Float>e

Richard A. Harmon
!


!Float commentStamp: '<historical>' prior: 0!
My instances represent IEEE-754 floating-point double-precision numbers.  They have about 16 digits of accuracy and their range is between plus and minus 10^307. Some valid examples are:
	
	8.0 13.3 0.3 2.5e6 1.27e-30 1.27e-31 -12.987654e12

Mainly: no embedded blanks, little e for tens power, and a digit on both sides of the decimal point.

Implementation;

	This is a simplified overview of Float representation.  A full description of the IEEE-754 standard is documented in many places on the world-wide web and elsewhere.  If any details below differ with IEEE-754 standard, it is correct and the description below is wrong.

    1 bit    - sign range: 1 or 0
				0 is positive and 1 is negative.
   11 bits   - signed exponent range: -1023 (-16r3FF) through +1024 (16r400)
				value of bits minus 1023 (16r3FF).
   52 bits   - normalized mantissa range: 0 through 16r1FFFFFFFFFFFFF
				value of bits plus an assumed high order 1 bit.

The following are reserved values:

	Reserved			Sign Exponent		Mantissa   Accessing
	Class var. name			 raw bits		raw bits
	-----------------------	----	------------------	------------ --------------------
    	* (zero)				+	 -1023 (16r0)		0		   0.0
    	NegativeZero		-	   "      "		0		   Float negativeZero
  	Infinity				+	 1024 (16r7FF)	0		   Float infinity
	NegativeInfinity	-	   "      "		0		   0.0 - Float infinity
   	NaN					+	 1024 (16r7FF)	16r8000000000000		Float nan
	* (denormalized)		+	 1024 (16r7FF)	not 0	   -


The following are a couple of special values:

	Special		Sign Exponent		Mantissa   			Accessing (ANSI)
	------------	----- -----------------	-----------------------	--------------------
???	Epsilon		+	-40 (16r3D7)		16r19799812DEA11		Float epsilon
	MaxVal		+	 1023 (16r7FE)	16rFFFFFFFFFFFFF		Float fmax

Notes:

Largest value that can be stored in n bits is:

	((2 raisedTo: n) - 1).

Exponent offset is:

	((Largest value - sign bit) / 2) negated

Largest signed exponent that can be stored in n bits is:

	(Largest value) - (Exponent offset)
!

Date subclass: #RuleDate
	instanceVariableNames: 'dayOfWeek selectionRule '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Magnitudes'!

!RuleDate commentStamp: '<historical>' prior: 0!
RuleDate instances represent dates determined by some rule (first Tuesday in March, etc.) but the exact day of the month varies from year to year.  Once created they must be updated for a selected year to represent the exact day of that year.

Typical Use:

	RuleDate instances are suitable for representing dates such as election day, or the start of daylight saving time in the USA.

Implementation:

Instance variables:

	dayOfWeek		- a <Symbol> representing the day name of the
					week (Sunday, etc) of the desired day.
	selectionRule	- a <Symbol> representing the position (first, last)
					of the desired day in the list of days occurring
					on that day of the week in the desired month.
!

RuleDate subclass: #RuleIndexDate
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Magnitudes'!

!RuleIndexDate commentStamp: '<historical>' prior: 0!
RuleIndexDate instances represent dates determined by some rule but the exact day of the month varies from year to year.  Once created they must be updated for a selected year to represent the exact day of that year.
	It has a rule that selects exact day of the month based on the ordinal position (1, 2, etc.) of the desired day in the list of days occurring on that day of the week in the desired month.

Typical Use:

	RuleIndexDate instances are suitable for representing dates such as Thanksgiving Day (fourth Tuesday in November) in the USA.

Implementation:

Instance variables:

	(selectionRule in super class)	- an <Integer> representing the
			ordinal position (1, 2, etc.) of the desired day in the list
			of days occurring on that day of the week in the desired
			month.
!

RuleDate subclass: #RuleSelectionCodeDate
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Magnitudes'!

!RuleSelectionCodeDate commentStamp: '<historical>' prior: 0!
RuleSelectionCodeDate instances represent dates determined by some rule but the exact day of the month varies from year to year.  Once created they must be updated for a selected year to represent the exact day of that year.
	It has a rule that selects exact day of the month by evaluating the selection block with the list of days occurring on that day of the week in the desired month as an argument.

Typical Use:

	RuleSelectionCodeDate instances are suitable for representing dates such as the last day of daylight saving time (the Saturday before the last Sunday in October) in the USA.

Implementation:

Instance variables:

	(selectionRule in super class)	- a <monadicValuable> containing
			the selection block to be evaluated.  The argument is the list
			of days occurring on that day of the week in the desired
			month.  It must return the desired day of the month 			which may be any <Integer>.
!

Number subclass: #ScaledDecimal
	instanceVariableNames: 'fraction scale '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Numbers'!

!ScaledDecimal commentStamp: '<historical>' prior: 0!
ScaledDecimal provides a numeric representation of fixed point decimal numbers able to accurately represent decimal fractions.  It supports unbounded precision, with no limit to the number of digits before and after the decimal point.
	ScaledDecimal complies with the ANSI protocols:

	Object
	number
	ScaledDecimal
	????
#todo. "finish protocol list."

Implementation Notes:

	I implemented ScaledDecimal with the decimal fraction stored in instance variables numerator and denominator, and the number of digits after the decimal point in fractionalDigit as a positive Integer.  I implemented operations by first coercing the aurguments to fractions, doing the operations, then coercing the result to the proper numeric representation when necessary and scale.  This is because I assume the Fraction class is more likely to implement them correctly.  

Richard A. Harmon
!


!Object methodsFor: 'error handling' stamp: 'ls 7/5/2000 15:39'!
doesNotUnderstand: aMessage 
	"Handle the fact that there was an attempt to send the given message to 
	the receiver but the receiver does not understand this message 
	(typically sent from the machine when a message is sent to the 
	receiver and no method is defined for that selector)."


	MessageNotUnderstood new message: aMessage;
	 signal.
	^ aMessage sentTo: self! !

!Object methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isScaledDecimal
	"Reimplementation - Number 'testing' method."
	^ false! !


!BlockContext methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
argumentCount
	"Answers the number of arguments needed to evaluate the receiver."
	#Valuable.
	^ self numArgs! !


!Character methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
codePoint
	"Return the encoding value of the receiver."
	#Fundmntl.
	^ self asciiValue! !

!Character methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isCharacter
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	^ true! !


!Character class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
codePoint: integer 
	"Return a character whose encoding value is integer."
	#Fundmntl.
	(0 > integer or: [255 < integer])
		ifTrue: [self error: 'parameter out of range 0..255'].
	^ CharacterTable at: integer + 1! !


!Collection methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asArray
	"Answer an Array whose elements are the elements of the receiver."
	| newArray index |
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> protocol."
	newArray := Array new: self size.
	index := 1.
	self
		do: 
			[:element | 
			newArray at: index put: element.
			index := index + 1].
	^ newArray! !

!Collection methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asByteArray
	"Answer a byte array whose elements are the elements of the receiver."
	| answer index |
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> protocol."
	(self allSatisfy: [:element | ByteArray meetsElementTypeConstraints: element])
		ifFalse: [^ ByteArray errorElementType].
	answer := ByteArray new: self size.
	index := 1.
	self
		do: 
			[:element | 
			answer at: index put: element.
			index := index + 1].
	^ answer! !

!Collection methodsFor: 'math functions' stamp: 'RAH 4/25/2000 19:49'!
allSatisfy: discriminator 
	"Return true if the discriminator evaluates to true for every element of 
	the receiver, else false.  Note: Return true if the receiver is empty."
	#SIFMod.
	"Changed 200/02/12 Previous verion can produce 
	
										huge temporary collections."
	self isEmpty ifTrue: [^ true].
	" 
	
    ^((self reject: discriminator 
	
    ) size > 0) not 
	
"
	self
		do: [:elem | (discriminator value: elem)
				ifFalse: [^ false]].
	^ true! !

!Collection methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
notEmpty
	"Return true if the receiver contains elements.  Return false otherwise."
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> protocol."
	^ self isEmpty not! !


!Bag methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
setDictionary: anInteger 
	"Private - Initialize the receiver for anInteger constrained number of 
	elements. "
	#Collectn.
	"Added 2000/04/08 For ANSI <Bag factory> support."
	contents := Dictionary new: anInteger! !


!Collection class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
meetsElementTypeConstraints: element 
	"Private - Answer a Boolean, true if element meets the element type 
	constraints of the receiver collection, false otherwise."
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	^ true! !


!Bag class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
new: countInteger 
	"Create a new collection.  The parameter countInteger constrains the 
	number of elements in the result."
	#Collectn.
	"Added 2000/04/08 For ANSI <Bag factory> protocol."
	^ super new setDictionary: countInteger! !


!Date methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isDate
	#DtAndTm.
	"Added 2000/04/08 To use DNU mod."
	^ true! !


!DateAndTime methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asBC
	"Answer a <DateAndTime> the receiver interpreted as BC."
	| bcDate |
	bcDate := Date newDay: localDate day year: 1 - localDate year.
	^ self class
		newRationalLocalDate: bcDate
		time: localTime
		offset: localOffsetToUTC
		timeZoneName: localTimeZoneName
		timeZoneAbbrev: localTimeZoneAbbrev! !

!DateAndTime methodsFor: 'converting' stamp: 'ls 6/30/2000 11:30'!
asLocal
	"Answer a <DateAndTime> that represents the same UTC time as the 
	receiver, but in the local time specified by the implementation."
	| referenceTime |
	self flag: #xxx. "this implementation assumes that the timezone currently in effect should be used on the receiver.  This causes funny effects, but should be 'correct' (whatever 'correct' means for local times)"
	referenceTime := DateAndTime now.
	^ self class
		newRationalLocalDate: localDate
		time: localTime + referenceTime offset
		offset: referenceTime offset
		timeZoneName: referenceTime timeZoneName
		timeZoneAbbrev: referenceTime timeZoneAbbreviation! !

!DateAndTime methodsFor: 'converting' stamp: 'ls 6/30/2000 11:25'!
asUTC
	"Answer a <DateAndTime> that represents the same absolute time as the 
	receiver but in the local time UTC."
	^ self class
		newRationalLocalDate: localDate
		time: localTime - self offset
		offset: Duration zero
		timeZoneName: 'Universal Coordinated Time'
		timeZoneAbbrev: 'UTC'! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
dayOfMonth
	"Answer an <integer> between 1 and 31, inclusive, representing the day 
	of the month in the local time of the receiver, which includes the 
	receiver. "
	^ localDate dayOfMonth! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
dayOfWeek
	"Answer an <integer> between 1 and 7, inclusive, representing the day of 
	the week in the local time of the receiver, which includes the receiver. 
	 Sunday is 1, Monday 2, and so on."
	| dayIndex |
	dayIndex := localDate weekdayIndex + 1.
	dayIndex > 7 ifTrue: [dayIndex := 1].
	^ dayIndex! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
dayOfWeekAbbreviation
	"Answer a <readableString> which is the abbreviated  name of the day 
	of the week in the local time of the receiver, which includes the 
	receiver. "
	^ self dayOfWeekName copyFrom: 1 to: 3! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
dayOfWeekName
	"Answer a <readableString> which is the name of the day of the week in 
	the local time of the receiver, which includes the receiver."
	^ localDate weekday asString! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
dayOfYear
	"Answer an <integer> between 1 and 366, inclusive, representing the day 
	of the year in the local time of the receiver, which includes the 
	receiver. "
	^ localDate day! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
hour
	"Answer an <integer> between 0 and 23, inclusive, representing the hour 
	of the day in the local time of the receiver.  This may be in either the 
	12 or 24 hour clock."
	^ self localTime hours! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
hour12
	"Answer an <integer> between 1 and 12, inclusive, representing the hour 
	of the day in the 12-hour clock of the local time of the receiver."
	| tempHour |
	tempHour := (self hour + 1) abs.
	tempHour > 12 ifTrue: [tempHour := tempHour - 12].
	^ tempHour! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
hour24
	"Answer an <integer> between 0 and 23, inclusive, representing the hour 
	of the day in the 24-hour clock of the local time of the receiver."
	^ self hour! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
meridianAbbreviation
	"Answer a <readableString> which is the abbreviated name of the half of 
	the day in the local time of the receiver, which includes the receiver."
	^ self class amOrPmAbbrevForHourIn24: self hour24! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
minute
	"Answer an <integer> between 0 and 59, inclusive, representing the 
	minute of the hour in the local time of the receiver."
	^ self localTime minutes! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
month
	"Answer an <integer> between 1 and 12, inclusive, representing the 
	month of the year in the local time of the receiver."
	^ localDate monthIndex! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
monthAbbreviation
	"Answer a <readableString> which is the abbreviated  name of the month 
	of the year in the local time of the receiver, which includes the 
	receiver. "
	^ self monthName copyFrom: 1 to: 3! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
monthName
	"Answer a <readableString> which is the name of the month of  the year 
	in the local time of the receiver, which includes the receiver."
	^ localDate monthName asString! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
offset
	"Answer a <Duration> which is the difference between the local time of 
	the receiver, and UTC at the time of the receiver."
	^ localOffsetToUTC! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
offset: offsetDuration 
	"Answer a <DateAndTime> equivalent to the receiver, but with its local 
	time offset from UTC by the <Duration>, offset."
	self class confirmOffset: offsetDuration.
	^ self class
		newRationalLocalDate: localDate
		time: localTime
		offset: offsetDuration
		timeZoneName: localTimeZoneName
		timeZoneAbbrev: localTimeZoneAbbrev! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
second
	"Answer a <number> in the range 0 <= s < 60, representing the second of 
	the minute of the local time of the receiver, including any fractional 
	part. "
	^ self localTime seconds abs! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
timeZoneAbbreviation
	"Answer the abbreviation of the name, in the local time of the receiver, 
	of the time zone of the receiver."
	^ localTimeZoneAbbrev! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
timeZoneName
	"Answer the name in the local time of the receiver of the time zone of 
	the receiver."
	^ localTimeZoneName! !

!DateAndTime methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
year
	"Answer an <integer> representing the year of the local time which 
	includes the receiver."
	^ localDate year! !

!DateAndTime methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
localDate
	"Private - Answer a Date, the date in local time of the receiver."
	^ localDate! !

!DateAndTime methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
localTime
	"Private - Answer a <Duration>, the local time of the receiver."
	^ localTime! !

!DateAndTime methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
rationaliseLocalTime
	"Private - Adjust the receiver so localTime is less than a day and 
	positive. "
	"Insure time is positive."
	| timeDays |
	localTime negative
		ifTrue: 
			[timeDays := localTime days abs + 1.
			localDate := localDate subtractDays: timeDays.
			localTime := (Duration days: timeDays)
						+ localTime].
	"Insure time is less than a day."
	timeDays := localTime days.
	timeDays > 0
		ifTrue: 
			[localDate := localDate addDays: timeDays.
			localTime := localTime - (Duration days: timeDays)]! !

!DateAndTime methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
setLocalDate: localDateIn time: localTimeIn offset: localOffsetIn timeZoneName: tzNameIn timeZoneAbbrev: tzAbbrevIn 
	"Private - Set the initial values."
	localDate := localDateIn.
	localTime := localTimeIn.
	localOffsetToUTC := localOffsetIn.
	localTimeZoneName := tzNameIn.
	localTimeZoneAbbrev := tzAbbrevIn! !

!DateAndTime methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
+ operandDuration 
	"Answer a <DateAndTime>, the result of adding the <Duration> argument, 
	operand to the receiver.  More specifically: Answer a <DateAndTime> 
	that represents the UTC time that is the <Duration> argument, operand, 
	after the receiver, and whose local time is the same as the receiver's.  
	If the argument is a negative <Duration>, then the result is a 
	<DateAndTime> prior to the receiver."
	Duration confirmDuration: operandDuration.
	^ self class
		newRationalLocalDate: localDate
		time: localTime + operandDuration
		offset: localOffsetToUTC
		timeZoneName: localTimeZoneName
		timeZoneAbbrev: localTimeZoneAbbrev! !

!DateAndTime methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
- operandDateAndTimeOrDuration 
	"Answer a <DateAndTime>, or a <Duration> that is the result of 
	subtracting the <DateAndTime>, or <Duration>, argument from the 
	receiver.  If the argument is a <DateAndTime>, then answer a <Duration> 
	whose value is the period of time between the argument and the 
	receiver.  The answer may be a negative <Duration> if the argument is 
	prior to the receiver in time."
	^ operandDateAndTimeOrDuration subtractFromDateAndTime: self! !

!DateAndTime methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
< operandDateAndTime 
	"Answer a <boolean>, whether the receiver is less than the <DateAndTime> 
	argument, operand, i.e. answer whether the receiver is prior to UTC 
	time represented by the argument."
	self class confirmDateAndTime: operandDateAndTime.
	self offset = operandDateAndTime offset
		ifTrue: 
			[localDate < operandDateAndTime localDate ifTrue: [^ true].
			(localDate = operandDateAndTime localDate and: [localTime < operandDateAndTime localTime])
				ifTrue: [^ true].
			^ false].
	^ self asUTC < operandDateAndTime asUTC! !

!DateAndTime methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
= comparand 
	"Answer a <boolean>, whether the argument is a <DateAndTime> 
	representing the same UTC time as the receiver.  The local times of the 
	receiver and the operand are ignored."
	| utcSelf utcComparand |
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	comparand isDateAndTime ifFalse: [^ false].
	utcSelf := self asUTC.
	utcComparand := comparand asUTC.
	^ utcSelf localDate = utcComparand localDate and: [utcSelf localTime = utcComparand localTime]! !

!DateAndTime methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
> operandDateAndTime 
	"Answer a <boolean>, whether the receiver is greater than the 
	<DateAndTime> argument, operand, i.e. answer whether the receiver is 
	after the UTC time represented by the argument."
	self class confirmDateAndTime: operandDateAndTime.
	self offset = operandDateAndTime offset
		ifTrue: 
			[localDate > operandDateAndTime localDate ifTrue: [^ true].
			(localDate = operandDateAndTime localDate and: [localTime > operandDateAndTime localTime])
				ifTrue: [^ true].
			^ false].
	^ self asUTC > operandDateAndTime asUTC! !

!DateAndTime methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
hash
	"Answer an <integer>, a hash code that can be used in conjunction with 
	an #= comparison."
	| utcSelf |
	utcSelf := self asUTC.
	^ utcSelf localDate hash bitXor: utcSelf localTime hash! !

!DateAndTime methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isDateAndTime
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	^ true! !

!DateAndTime methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isLeapYear
	"Answer a <boolean>, whether the local time year which includes the 
	receiver is a leap year."
	^ (Date leapYear: self year)
		= 1! !

!DateAndTime methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isZero
	"Answer whether the receiver is zero."
	^ self = self class zero! !

!DateAndTime methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printOn: targetPuttableStream 
	"Write a textual description of the receiver to a stream. 
	The string of characters that would be the result of sending the 
	message #printString to the receiver is written to target."
	"Print -, if the year is < 0 otherwise space."
	| tempSeconds |
	self year < 0
		ifTrue: [targetPuttableStream nextPut: $-]
		ifFalse: [targetPuttableStream nextPut: Character space].
	"Print the year."
	targetPuttableStream nextPutAll: (self year abs printPaddedWith: $0 to: 4).
	targetPuttableStream nextPut: $-.
	"Print the month."
	targetPuttableStream nextPutAll: (self month printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $-.
	"Print the day of the month."
	targetPuttableStream nextPutAll: (self dayOfMonth printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $T.
	"Print the hour in the 24-hour clock."
	targetPuttableStream nextPutAll: (self hour24 printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $:.
	"Print the minute."
	targetPuttableStream nextPutAll: (self minute printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $:.
	"Print the second."
	targetPuttableStream nextPutAll: (self second printPaddedWith: $0 to: 2).
	"Print the offset."
	localOffsetToUTC >= Duration zero
		ifTrue: [targetPuttableStream nextPut: $+]
		ifFalse: [targetPuttableStream nextPut: $-].
	"Print the offset hours."
	targetPuttableStream nextPutAll: (localOffsetToUTC hours abs printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $:.
	"Print the offset minutes."
	targetPuttableStream nextPutAll: (localOffsetToUTC minutes abs printPaddedWith: $0 to: 2).
	"Print the offset seconds only if non-zero."
	tempSeconds := localOffsetToUTC seconds abs.
	tempSeconds ~= 0
		ifTrue: 
			[targetPuttableStream nextPut: $:.
			tempSeconds printOn: targetPuttableStream]! !

!DateAndTime methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printString
	"Answer a <readableString> representing the UTC time of the receiver 
	offset from UTC by the offset of the receiver.  All dates are in the 
	astronomical Gregorian calendar. 
	The result will be of the format: 
	 
	-YYYY-MM-DDThh:mm:ss.s+ZZ:zz:z 
	 
	where the z's specify the offset from UTC. e.g. 8:33:14.321 PM EST Jan 5, 
	1200 B.C. would be: 
	 
	-1199-01-05T20:33:14.321-05:00"
	| aStream |
	aStream := WriteStream on: (String new: 16).
	self printOn: aStream.
	^ aStream contents! !

!DateAndTime methodsFor: 'displaying' stamp: 'RAH 4/25/2000 19:49'!
displayOn: targetPuttableStream 
	"Append the end-user display representation of the receiver to target."
	self printOn: targetPuttableStream! !

!DateAndTime methodsFor: 'private-double dispatch' stamp: 'RAH 4/25/2000 19:49'!
subtractFromDateAndTime: aDateAndTime 
	"Private - Answer a <Duration>, the result of subtracting the receiver 
	from the known DateAndTime, aDateAndTime."
	^ Duration seconds: aDateAndTime localDate asSeconds - localDate asSeconds + (aDateAndTime localTime asSeconds - localTime asSeconds)! !


!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
checkMonth: month 
	"Private - Check the validity of the month number month.  If the month 
	is invalid, report the error."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	month isInteger ifFalse: [^ self error: 'month must be an integer.'].
	(month between: 1 and: 12)
		ifFalse: [^ Date errorInMonth]! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
checkYear: year dayOfYear: dayOfYear 
	"Private - Check the validity of the year and dayOfYear.  If invalid, 
	report an error."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	year isInteger ifFalse: [^ self error: 'year must be an integer.'].
	(dayOfYear isInteger and: [dayOfYear between: 1 and: 366])
		ifFalse: [^ self error: 'day must be an integer between 1 and 366 inclusive.']! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
checkYear: year month: month day: dayOfMonth 
	"Private - Check the validity of the year and dayOfYear.  If invalid, 
	report an error."
	| daysInMonth |
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	year isInteger ifFalse: [^ self error: 'year must be an integer.'].
	self checkMonth: month.
	dayOfMonth isInteger ifFalse: [^ self error: 'day must be an integer.'].
	daysInMonth := Date daysInMonth: month forYear: year.
	(dayOfMonth between: 1 and: daysInMonth)
		ifFalse: [^ self error: 'day must be between 1 and ' , daysInMonth printString , ' inclusive for month ' , month printString , ' of year ' , year printString , '.']! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmDate: date 
	"Private - Report an error if date is not an instances of Date."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	date isDate ifFalse: [^ self error: 'Operand does not represent a Date.']! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmDateAndTime: dateAndTime 
	"Private - Report an error if dateAndTime is not an instances of 
	DateAndTime. "
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	dateAndTime isDateAndTime ifFalse: [^ self error: 'DateAndTime & ' , dateAndTime class name , ' are not comparable.']! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmOffset: durationObject 
	"Private - Report an error if durationObject does not represent a 
	Duration less than one day."
	Duration confirmDuration: durationObject.
	durationObject abs < (Duration days: 1) ifFalse: [^ self error: 'Not in range: -24 < offset < +24 hours.']! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmRuleDate: ruleDate 
	"Private - Report an error if ruleDate is not an instances of RuleDate."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	ruleDate isRuleDate ifFalse: [^ self error: 'Operand does not represent a RuleDate.']! !

!DateAndTime class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmString: string 
	"Private - Report an error if string is not an instances of String."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	string isString ifFalse: [^ self error: 'Operand does not represent a String.']! !

!DateAndTime class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
year: yearInteger day: dayOfYearInteger hour: hourInteger minute: minuteInteger second: secondInteger 
	"Answer a <DateAndTime> which is the <number> second, of the <integer> 
	minute, of the <integer> hour, of the <integer> day of the year, 
	dayOfYear, of the <integer> year, of the astronomical Gregorian 
	calendar offset from UTC by the <Duration> offset, where: 
	 
	1 <= day <= 366, and year can be any <integer>, 
	0 <= second < 60, 0 <= minute <= 59, 0 <= hour <= 23 
	 
	If the specified time does not exist in the local time defined by the 
	implementation the result is the least <DateAndTime> that conforms to the 
	given parameters."
	^ self
		year: yearInteger
		day: dayOfYearInteger
		hour: hourInteger
		minute: minuteInteger
		second: secondInteger
		offset: self defaultOffsetFromUTC! !

!DateAndTime class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
year: yearInteger day: dayOfYearInteger hour: hourInteger minute: minuteInteger second: secondInteger offset: offsetDuration 
	"Answer a <DateAndTime> which is the local time second, second, of the 
	minute, minute, of the hour, hour, of the day dayOfYear of the year 
	year of the astronomical Gregorian calendar offset from UTC by offset."
	| localDate localTimeDuration |
	self checkYear: yearInteger dayOfYear: dayOfYearInteger.
	localDate := Date newDay: dayOfYearInteger year: yearInteger.
	localTimeDuration := Duration
				days: 0
				hours: hourInteger
				minutes: minuteInteger
				seconds: secondInteger.
	Duration confirmDuration: offsetDuration.
	^ self
		newRationalLocalDate: localDate
		time: localTimeDuration
		offset: offsetDuration
		timeZoneName: self defaultTimeZoneName
		timeZoneAbbrev: self defaultTimeZoneAbbrev! !

!DateAndTime class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
year: yearInteger month: monthInteger day: dayOfMonthInteger hour: hourInteger minute: minuteInteger second: secondInteger 
	"Answer a <DateAndTime> which is the <number> second, of the <integer> 
	minute, of the <integer> hour, of the <integer> day, dayOfMonth, of the 
	<integer> month, of the <integer> year, of the astronomical Gregorian 
	calendar in local time, where: 
	 
	0 <= second < 60, 0 <= minute <= 59, 0 <= hour <= 23, 1 <= day <= 31, 
	1 <= month <= 12, and year can be any <integer>.  
	 
	If the specified time does not exist in the local time defined by the 
	implementation the result is the least <DateAndTime> that conforms to the 
	given parameters."
	^ self
		year: yearInteger
		month: monthInteger
		day: dayOfMonthInteger
		hour: hourInteger
		minute: minuteInteger
		second: secondInteger
		offset: self defaultOffsetFromUTC! !

!DateAndTime class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
year: yearInteger month: monthInteger day: dayOfMonthInteger hour: hourInteger minute: minuteInteger second: secondInteger offset: offsetDuration 
	"Answer a <DateAndTime> which is the <number> second, of the <integer> 
	minute, of the <integer> hour, of the <integer> day, dayOfMonth, of the 
	<integer> month, of the <integer> year, of the astronomical Gregorian 
	calendar offset from UTC by the <Duration> offset, where: 
	 
	1 <= month <= 12, and year can be any <integer>,  
	0 <= second < 60, 0 <= minute <= 59, 0 <= hour <= 23, 1 <= day <= 31. 
	 
	If the specified time does not exist in the local time defined by the 
	implementation the result is the least <DateAndTime> that conforms to the 
	given parameters."
	| localDate localTimeDuration |
	self
		checkYear: yearInteger
		month: monthInteger
		day: dayOfMonthInteger.
	localDate := Date
				newDay: dayOfMonthInteger
				month: monthInteger
				year: yearInteger.
	"Chg Sqk"
	localTimeDuration := Duration
				days: 0
				hours: hourInteger
				minutes: minuteInteger
				seconds: secondInteger.
	Duration confirmDuration: offsetDuration.
	^ self
		newRationalLocalDate: localDate
		time: localTimeDuration
		offset: offsetDuration
		timeZoneName: self defaultTimeZoneName
		timeZoneAbbrev: self defaultTimeZoneAbbrev! !

!DateAndTime class methodsFor: 'private-constants' stamp: 'RAH 4/25/2000 19:49'!
amAbbrev
	"Private - Answer a <readableString>, the abbreviation of the half of the 
	day before noon."
	^ 'AM'! !

!DateAndTime class methodsFor: 'private-constants' stamp: 'RAH 4/25/2000 19:49'!
amOrPmAbbrevForHourIn24: hourInteger 
	"Private - Answer a <readableString>, the abbreviation of the half of the 
	day before or after noon in which the hour, hourInteger, falls in a 
	24-hour clock."
	hourInteger < 12 ifTrue: [^ self amAbbrev].
	^ self pmAbbrev! !

!DateAndTime class methodsFor: 'private-constants' stamp: 'RAH 4/25/2000 19:49'!
pmAbbrev
	"Private - Answer a String, the abbreviation of the half of the day 
	after noon."
	^ 'PM'! !

!DateAndTime class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
clockPrecision
	"Answer a <Duration> such that after that period of time passes, #now is 
	guaranteed to answer a different <DateAndTime>. Ideally this should be 
	the minimum such duration."
	#todo.
	"???Implement Find out what clockPrecision value is on WinTel. ???"
	^ Duration seconds: 1! !

!DateAndTime class methodsFor: 'constants' stamp: 'ls 6/30/2000 11:42'!
now
	"Answer a <DateAndTime> representing the current date and time."
	| timeArray localDate utcDate localTime utcTime seconds date time |
	timeArray := self primCurrentTime.
	timeArray ifNil: [
		"use the old primitive, and assume UTC time"
		seconds := self primSecondsClock.
		date := Date fromSeconds: seconds.
		time := Time fromSeconds: seconds \\ 86400. 
		timeArray := {
			date year.
			date dayOfYear.
			time hours.
			time minutes.
			time seconds.
			date year.
			date dayOfYear.
			time hours.
			time minutes.
			time seconds.
			'Universal Coordinated Time'.
			'UTC' }.
	].

	localDate := Date
		newDay: (timeArray at: 2)
		year: (timeArray at: 1).
	utcDate := Date
		newDay: (timeArray at: 7)
		year: (timeArray at: 6).

	localTime := Time new
		hours: (timeArray at: 3)
		minutes: (timeArray at: 4)
		seconds: (timeArray at: 5).

	utcTime := Time new
		hours: (timeArray at: 8)
		minutes: (timeArray at: 9)
		seconds: (timeArray at: 10).

	^ self
		newRationalLocalDate: (localDate)
		time: (Duration seconds: localTime asSeconds)
		offset: (Duration seconds: (localDate asSeconds + localTime asSeconds - utcDate asSeconds - utcTime asSeconds))
		timeZoneName: (timeArray at: 11)
		timeZoneAbbrev: (timeArray at: 12) ! !

!DateAndTime class methodsFor: 'initializing' stamp: 'ls 6/30/2000 10:08'!
initialize
	"Private - Initialize the class variables that describe constant values.  
	This initializes Duration first. 
	Note: Before returning the default offset, it, the default time zone name 
	and abbreviation are updated if todays date is during the daylight 
	saving period. 
	 
	DateAndTime initialize 
	."
	| curDate daylightStart daylightEnd |
	#todo.
	"??? Implement initialization of these variables from the operating system 
	environent settings. ???"
	"Duration must be initialized first:"
	Duration initialize.
	"Initialize the local offset, time zone name and abbreviation, 
	and Daylight Time start and end dates:"
	curDate := Date dateAndTimeNow at: 1.
[	daylightStart := RuleDate
				last: 'Sunday'
				inMonth: 'April'
				year: curDate year.
	daylightEnd := RuleSelectionCodeDate
				selected: 'Sunday'
				inMonth: 'October'
				year: curDate year
				byCode: [:sundaysList | "Standard Time start, so Daylight Time end is previous 
					day: "
					sundaysList last - 1].
	self
		setStandardOffset: (Duration hours: -6)
		standardTimeZoneName: 'Central Standard Time'
		standardTimeZoneAbbrev: 'CST'
		daylightOffset: (Duration hours: -5)
		daylightTimeZoneName: 'Central Daylight Time'
		daylightTimeZoneAbbrev: 'CDT'
		daylightStartDate: daylightStart
		daylightEndDate: daylightEnd.
].	DefaultIsStandard := true.
	self updateDefaultOffsetEtcIfRequired! !

!DateAndTime class methodsFor: 'private-accessing' stamp: 'RAH 4/25/2000 19:49'!
defaultOffsetFromUTC
	"Private - Answer a <Duration>, the default local offset from UTC. 
	Note: The default offset, time zone name and abbreviation are switched 
	to daylight saving when today is the start of daylight saving and to 
	standard when today is the end of daylight saving."
	self updateDefaultOffsetEtcIfRequired.
	DefaultIsStandard ifFalse: [^ DaylightOffset].
	^ StandardOffset! !

!DateAndTime class methodsFor: 'private-accessing' stamp: 'RAH 4/25/2000 19:49'!
defaultTimeZoneAbbrev
	"Private - Answer a <String>, the default local time zone abbreviation. 
	Note: The default offset, time zone name and abbreviation are switched 
	to daylight saving when today is the start of daylight saving and to 
	standard when today is the end of daylight saving."
	DefaultIsStandard ifFalse: [^ DaylightTimeZoneAbbrev].
	^ StandardTimeZoneAbbrev! !

!DateAndTime class methodsFor: 'private-accessing' stamp: 'RAH 4/25/2000 19:49'!
defaultTimeZoneName
	"Private - Answer a <String>, the default local time zone name. 
	Note: The default offset, time zone name and abbreviation are switched 
	to daylight saving when today is the start of daylight saving and to 
	standard when today is the end of daylight saving."
	DefaultIsStandard ifFalse: [^ DaylightTimeZoneName].
	^ StandardTimeZoneName! !

!DateAndTime class methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
newRationalLocalDate: localDate time: localTime offset: localOffset timeZoneName: tzName timeZoneAbbrev: tzAbbrev 
	"Private - Answer a <DateAndTime> which is the local time, localTime, of 
	the local date, localDate, of the astronomical Gregorian calendar offset 
	from UTC by localOffset with the local time zone name, tzName, and 
	abbreviation, tzAbbrev. 
	Note: This is the normal internal instance creation method for 
	operations such as #+, etc. where we know we are working in local 
	time. "
	| newDateAndTime |
	newDateAndTime := self new.
	newDateAndTime
		setLocalDate: localDate
		time: localTime
		offset: localOffset
		timeZoneName: tzName
		timeZoneAbbrev: tzAbbrev.
	newDateAndTime rationaliseLocalTime.
	^ newDateAndTime! !

!DateAndTime class methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
setStandardOffset: standardOffset standardTimeZoneName: standardTZName standardTimeZoneAbbrev: standardTZAbbrev daylightOffset: daylightOffset daylightTimeZoneName: daylightTZName daylightTimeZoneAbbrev: daylightTZAbbrev daylightStartDate: daylightStart daylightEndDate: daylightEnd 
	"Set the users standard offset from UTC, time zone name and 
	abbreviation; daylight saving offset, name and abbreviation; and the 
	daylight saving period start and end (day BEFORE start of standard time) 
	dates. Note: Before returning the default offset, it, the default time zone 
	name and abbreviation are updated if todays date is during the 
	daylight saving period."
	self confirmOffset: standardOffset.
	StandardOffset := standardOffset.
	self confirmString: standardTZName.
	StandardTimeZoneName := standardTZName.
	self confirmString: standardTZAbbrev.
	StandardTimeZoneAbbrev := standardTZAbbrev.
	self confirmOffset: daylightOffset.
	DaylightOffset := daylightOffset.
	self confirmString: daylightTZName.
	DaylightTimeZoneName := daylightTZName.
	self confirmString: daylightTZAbbrev.
	DaylightTimeZoneAbbrev := daylightTZAbbrev.
	self confirmRuleDate: daylightStart.
	DaylightStartDate := daylightStart.
	self confirmRuleDate: daylightEnd.
	DaylightEndDate := daylightEnd! !

!DateAndTime class methodsFor: 'private-utility' stamp: 'ls 6/30/2000 10:09'!
updateDefaultOffsetEtcIfRequired
	"Private - Updated the default offset, time zone name and abbreviation 
	when they differ from todays date.  When todays date is the start of 
	daylight saving switch default to local daylight saving offset, time zone 
	name and abbreviation.  When the start of standard switch default to 
	local standard."
	| curDate |
true ifTrue: [ ^self ].
	curDate := Date dateAndTimeNow at: 1.
	curDate year = DaylightStartDate year
		ifFalse: 
			[DaylightStartDate updateForYear: curDate year.
			DaylightEndDate updateForYear: curDate year.
			DefaultIsStandard := true].
	(curDate between: DaylightStartDate and: DaylightEndDate)
		ifTrue: [DefaultIsStandard ifTrue: [DefaultIsStandard := false]]
		ifFalse: ["Currently in Standard time."
			DefaultIsStandard ifFalse: [DefaultIsStandard := true]]! !


!Dictionary methodsFor: 'adding' stamp: 'RAH 4/25/2000 19:49'!
addAll: dictionary 
	"Store the elements of dictionary in the receiver at the corresponding 
	keys from dictionary."
	#Collectn.
	"Added 2000/04/08 For ANSI <abstractDictionary> protocol."
	dictionary keysAndValuesDo: [:key :value | self at: key put: value]! !

!Dictionary methodsFor: 'removing' stamp: 'RAH 4/25/2000 19:49'!
removeAllKeys: keys 
	"Remove any elements from the receiver which are stored at the keys 
	specified in keys."
	#Collectn.
	"Added 2000/04/08 For ANSI <abstractDictionary> protocol."
	keys do: [:aKey | self removeKey: aKey]! !

!Dictionary methodsFor: 'removing' stamp: 'RAH 4/25/2000 19:49'!
removeAllKeys: keys ifAbsent: operationMonadicValuable 
	"Remove any elements from the receiver which are stored at the keys 
	specified in keys.  For any element in keys which is not a valid key of 
	the receiver, evaluate operationMonadicValuable with that element as 
	the argument, but do not stop the enumeration."
	#Collectn.
	"Added 2000/04/08 For ANSI <abstractDictionary> protocol."
	keys do: [:aKey | self removeKey: aKey ifAbsent: [operationMonadicValuable value: aKey]]! !

!Dictionary methodsFor: 'enumerating' stamp: 'RAH 4/25/2000 19:49'!
collect: transformer 
	"Answer a new instance of the receiver's type with the same keys.  For 
	each key of the answer, a new element is obtained by evaluating 
	transformer with the corresponding element of the receiver as the 
	parameter. "
	| answer |
	#Collectn.
	"Changed 2000/04/08 For ANSI <abstractDictionary> protocol."
	answer := self species new.
	"<- Chg."
	self keysAndValuesDo: [:key :value | "<- Chg."
		"<- Chg."
		answer at: key put: (transformer value: value)].
	^ answer! !

!Dictionary methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asSet
	"Answer a set with the same elements as the receiver."
	| aSet |
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> protocol."
	aSet := Set new: self size.
	self do: [:each | aSet add: each].
	^ aSet! !


!Duration methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
days
	"Answer the the <integer> number of whole days represented by the 
	receiver.  The answer may be negative for negative durations."
	^ self asSeconds abs // SecondsInDay * self asSeconds sign! !

!Duration methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
hours
	"Answer an <integer> beteen -23 and 23, inclusive, which is the number 
	of whole hours represented by the receiver after complete days have 
	been removed."
	^ self asSeconds abs \\ SecondsInDay // SecondsInHour * self asSeconds sign! !

!Duration methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
minutes
	"Answer an <integer> beteen -59 and 59, inclusive, which is the number 
	of whole minutes represented by the receiver after complete days and 
	hours have been removed."
	^ self asSeconds abs \\ SecondsInDay \\ SecondsInHour // SecondsInMinute * self asSeconds sign! !

!Duration methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
seconds
	"Answer a <number> in the range -60 < s < 60, which is the number of 
	seconds, including any fraction part, represented by the receiver after 
	complete days, hours and minutes have been removed."
	^ self asSeconds abs \\ SecondsInDay \\ SecondsInHour \\ SecondsInMinute * self asSeconds sign! !

!Duration methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
* operandNumber 
	"Answer the result of multiplying the receiver by the <number> 
	argument, operand."
	^ Duration seconds: self asSeconds * operandNumber! !

!Duration methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
+ operandDuration 
	"Answer the result of adding the <Duration> argument, operand to the 
	receiver. "
	self class confirmDuration: operandDuration.
	^ Duration seconds: self asSeconds + operandDuration asSeconds! !

!Duration methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
- operandDuration 
	"Answer a new <Duration> that is the the result of subtracting the 
	<Duration> argument from the receiver."
	self class confirmDuration: operandDuration.
	^ Duration seconds: self asSeconds - operandDuration asSeconds! !

!Duration methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
/ operandNumberOrDuration 
	"Answer the result of dividing the receiver by operand.  If the operand 
	is a <number> then the result will be a new <Duration>.  If, on the other 
	hand, the operand is a <Duration> then the result will be a <number>. 
	Raise a ZeroDivide exception if the operand is zero."
	self class confirmDurationOrNumber: operandNumberOrDuration.
	operandNumberOrDuration isZero ifTrue: [^ (ZeroDivide dividend: self) signal].
	^ operandNumberOrDuration divideIntoDuration: self! !

!Duration methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
abs
	"Answer a <number> that is the absolute value (positive magnitude) of 
	the receiver."
	^ Duration seconds: self asSeconds abs! !

!Duration methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
negated
	"Answer a <Duration> of the opposite sign as the receiver."
	^ Duration seconds: self asSeconds negated! !

!Duration methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printOn: targetPuttableStream 
	"Append a default string representation of the receiver to the 
	<readableStream>, target.  The result will be of the format: 
	 
	-D:HH:MM:SS[.S]"
	"Print the minus sign, if required."
	self asSeconds negative ifTrue: [targetPuttableStream nextPut: $-].
	"Print the absolute value of the receiver."
	self printAbsOn: targetPuttableStream! !

!Duration methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printString
	"Answer a description of the receiver that is formatted as 
	[-]D:HH:MM:SS[.S] where: 
	 
	-	is a minus sign if the receiver represents a length of time going from 
	the future into the past, 
	D	is the number of complete days with leading zeros to fill one place, 
	HH	is the number of complete hours with leading zeros to fill two places, 
	MM	is the number of complete minutes with leading zeros to fill two 
	places, SS	is. the number of complete seconds with leading zeros to fill 
	two places, and 
	.S	is the fractional part of the number of seconds, if any."
	| aStream |
	aStream := WriteStream on: (String new: 16).
	self printOn: aStream.
	^ aStream contents! !

!Duration methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isDuration
	#DtAndTm.
	"Added 2000/04/08 To use DNU mod."
	^ true! !

!Duration methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isZero
	"Answer whether the receiver is zero."
	^ self = self class zero! !

!Duration methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
negative
	"Answer whether the receiver is negative."
	^ self asSeconds negative! !

!Duration methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
positive
	"Answer whether the receiver is positive or zero."
	^ self asSeconds positive! !

!Duration methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
asSeconds: aNumber 
	"Private - Set the total number of seconds in the length of time 
	represented by the receiver to aNumber."
	asSeconds := aNumber! !

!Duration methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
printAbsOn: targetPuttableStream 
	"Private - Append the ASCII representation of the absolute value of the 
	receiver. "
	"Print the days."
	| tempSeconds |
	targetPuttableStream nextPutAll: (self days abs printPaddedWith: $0 to: 1).
	targetPuttableStream nextPut: $:.
	"Print the hours."
	targetPuttableStream nextPutAll: (self hours abs printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $:.
	"Print the minutes."
	targetPuttableStream nextPutAll: (self minutes abs printPaddedWith: $0 to: 2).
	targetPuttableStream nextPut: $:.
	"Print the seconds."
	tempSeconds := self seconds abs.
	tempSeconds fractionPart = 0 ifTrue: [tempSeconds := tempSeconds asInteger].
	targetPuttableStream nextPutAll: (tempSeconds printPaddedWith: $0 to: 2)! !

!Duration methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
sign
	"Private - Answer 1 if the receiver is greater than zero, answer -1 if the 
	 receiver is less than zero, else answer zero."
	self > 0 ifTrue: [^ 1].
	self < 0 ifTrue: [^ -1].
	^ 0! !

!Duration methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asSeconds
	"Answer the <number> of seconds represented by the receiver, including 
	any fractional part."
	^ asSeconds! !

!Duration methodsFor: 'private-double dispatch' stamp: 'RAH 4/25/2000 19:49'!
divideIntoDuration: aDuration 
	"Private - Answer a <Duration>, the result of dividing the receiver into 
	the known Duration, aDuration."
	^ aDuration asSeconds / self asSeconds! !

!Duration methodsFor: 'private-double dispatch' stamp: 'RAH 4/25/2000 19:49'!
subtractFromDateAndTime: aDateAndTime 
	"Private - Subtract the receiver from the known DateAndTime, 
	aDateAndTime. "
	^ aDateAndTime + self negated! !

!Duration methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
< operandDuration 
	"Answer whether the receiver is less than the <Duration> argument, 
	operand (i.e. the receiver represents a shorter duration)."
	self class confirmDuration: operandDuration.
	^ self asSeconds < operandDuration asSeconds! !

!Duration methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
= comparand 
	"Answer whether the argument is a <Duration> representing the same 
	period of time as the receiver."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	comparand isDuration ifTrue: [^ self asSeconds = comparand asSeconds].
	^ false! !

!Duration methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
> operandDuration 
	"Answer whether the receiver is greater than the <Duration> argument, 
	operand (i.e. the receiver represents a longer duration)."
	self class confirmDuration: operandDuration.
	^ self asSeconds > operandDuration asSeconds! !

!Duration methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
hash
	"Answer the hash value for the receiver."
	^ self asSeconds hash! !

!Duration methodsFor: 'displaying' stamp: 'RAH 4/25/2000 19:49'!
displayOn: targetPuttableStream 
	"Append the end-user display representation of the receiver to target."
	self printOn: targetPuttableStream! !


!Duration class methodsFor: 'initializing' stamp: 'RAH 4/25/2000 19:49'!
initialize
	"Private - Initialize the class variables that describe constant values. 
	 
	Duration initialize.
	"
	SecondsInDay := 86400.
	SecondsInHour := 3600.
	SecondsInMinute := 60! !

!Duration class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
zero
	"Answer the receiver's representation of zero."
	^ self seconds: 0! !

!Duration class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
days: daysInteger 
	"Answer a <Duration> of the number of days."
	self confirmInteger: daysInteger arg: 'days'.
	^ self seconds: daysInteger * SecondsInDay! !

!Duration class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
days: daysInteger hours: hoursInteger minutes: minutesInteger seconds: secondsNumber 
	"Answer a new instance of the receiver that represents the period of 
	time composed from the specified number of <integer> days, <integer> 
	hours, <integer> minutes, and <number> seconds."
	self confirmInteger: daysInteger arg: 'days'.
	self confirmInteger: hoursInteger arg: 'hours'.
	self confirmInteger: minutesInteger arg: 'minutes'.
	self confirmNumber: secondsNumber.
	^ self seconds: daysInteger * SecondsInDay + (hoursInteger * SecondsInHour) + (minutesInteger * SecondsInMinute) + secondsNumber! !

!Duration class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
hours: hoursInteger 
	"Answer a new instance of the receiver that represents the period of 
	time of <integer> hours."
	self confirmInteger: hoursInteger arg: 'hours'.
	^ self seconds: hoursInteger * SecondsInHour! !

!Duration class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
minutes: minutesInteger 
	"Answer a new instance of the receiver that represents the period of 
	time of <integer> minutes."
	self confirmInteger: minutesInteger arg: 'minutes'.
	^ self seconds: minutesInteger * SecondsInMinute! !

!Duration class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
seconds: seconds 
	"Answer a new instance of the receiver that is <number> seconds long."
	| secondsFloatOrInteger |
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	self confirmNumber: seconds.
	seconds isInteger
		ifTrue: [secondsFloatOrInteger := seconds]
		ifFalse: [secondsFloatOrInteger := seconds asFloat].
	^ super new asSeconds: secondsFloatOrInteger! !

!Duration class methodsFor: 'private-utility' stamp: 'RAH 4/25/2000 19:49'!
confirmInteger: operand arg: argName 
	"Private - If operand does not represent an <integer> then signal an 
	error using argName."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	operand isInteger ifFalse: [^ self error: argName , ' does not represent a <integer>']! !

!Duration class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmDuration: operand 
	"Private - If operand does not represent a <Duration> then signal an 
	error. "
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	operand isDuration ifFalse: [^ self error: 'Operand does not represent a <Duration>']! !

!Duration class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmDurationOrNumber: operand 
	"Private - If operand does not represent a <Duration> or a <number> then 
	signal an error."
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	(operand isDuration or: [operand isNumber])
		ifFalse: [^ self error: 'Operand does not represent a <Duration> or a <number>']! !

!Duration class methodsFor: 'private-error checking' stamp: 'RAH 4/25/2000 19:49'!
confirmNumber: operand 
	"Private - If operand does not represent an <number> then signal an 
	error. "
	#DtAndTm.
	"Changed 2000/04/08 To use DNU mod."
	operand isNumber ifFalse: [^ self error: 'Operand does not represent a <number>']! !


!Number methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
\\ aNumber 
	"modulo. Remainder defined in terms of //. Answer a Number with the  
	same sign as aNumber. e.g. 9\\4 = 1, -9\\4 = 3, 9\\-4 = -3, 0.9\\0.4 = 0.1."
	#Numeric.
	"Changed 200/01/19 For ANSI to return correct type. 
	^26.0s3 - ((26.0s3 // (-3/2)) * (-3/2)) 
	^26.0s3 - (-18 * (-3/2)) 
	^26.0s3 - 27				returns -1.000s3 instead of the correct -1
	"
	aNumber isFraction ifTrue: ["<- Chg"
		^ self - (self // aNumber * aNumber) asFraction"<- Chg"].
	"<- Chg"
	^ self - (self // aNumber * aNumber)! !

!Number methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
raisedTo: operand 
	"Answer the receiver raised to the power operand. 
	Implementation: Caution - slightly different ANSI behavior (SDB)."
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	self = 0 & (operand <= 0) ifTrue: ["<- Chg"
		^ ArithmeticError signal: 'receiver = 0 and operand <= 0'].
	self < 0 ifTrue: ["<- Chg"
		^ ArithmeticError signal: 'receiver < 0'].
	operand isInteger ifTrue: [^ self raisedToInteger: operand].
	operand = 0 ifTrue: ["Special case of exponent = 0."
		^ 1].
	operand = 1 ifTrue: ["Special case of exponent = 1."
		^ self].
	"Otherwise raise it to the power using logarithms."
	^ (self asFloat ln * operand) exp! !

!Number methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
reciprocal
	"Answer 1 divided by the receiver. Create an error notification if the 
	receiver is 0."
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	self = 0 ifTrue: [^ (ZeroDivide dividend: self) signal"<- Chg"].
	^ 1 / self! !

!Number methodsFor: 'mathematical functions' stamp: 'RAH 4/25/2000 19:49'!
raisedToInteger: operand 
	"Answer the receiver raised to the power operand, an Integer."
	| count result |
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	operand isInteger ifFalse: [^ ArithmeticError signal: 'parameter is not an Integer'"<- Chg"].
	operand = 0 ifTrue: [^ self class one].
	operand = 1 ifTrue: [^ self].
	operand < 0 ifTrue: [^ (self raisedToInteger: operand negated) reciprocal].
	count := 1.
	[(count := count + count) < operand] whileTrue.
	result := self class one.
	[count > 0]
		whileTrue: 
			[result := result * result.
			(operand bitAnd: count)
				= 0 ifFalse: [result := result * self].
			count := count bitShift: -1].
	^ result! !

!Number methodsFor: 'truncation and round off' stamp: 'RAH 4/25/2000 19:49'!
fractionPart
	"Answer the fractional part of the receiver."
	#Numeric.
	"2000/03/04  Harmon R. Added ANSI <number> protocol"
	^ self - self truncated! !

!Number methodsFor: 'truncation and round off' stamp: 'RAH 4/25/2000 19:49'!
integerPart
	"Answer the integer part of the receiver."
	#Numeric.
	"2000/03/04  Harmon R. Added ANSI <number> protocol"
	^ self truncated! !

!Number methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
adaptToScaledDecimal: receiverScaledDecimal andSend: arithmeticOpSelector 
	"Do any required conversion and then the arithmetic. 
	receiverScaledDecimal arithmeticOpSelector self."
	#Numeric.
	"add 200/01/19 For ScaledDecimal support."
	^ self subclassResponsibility! !

!Number methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asFloatD
	"Answer a d precision floating-point number approximating the receiver."
	#Numeric.
	"add 200/01/19 For ANSI <number> protocol."
	^ self asFloat! !

!Number methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asFloatE
	"Answer a floating-point number approximating the receiver."
	#Numeric.
	"add 200/01/19 For ANSI <number> protocol."
	^ self asFloat! !

!Number methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asFloatQ
	"Answer a floating-point number approximating the receiver."
	#Numeric.
	"add 200/01/19 For ANSI <number> protocol."
	^ self asFloat! !

!Number methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asInteger
	"Answer the result of sending #rounded to the receiver."
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	^ self rounded"<- Chg: truncated"! !

!Number methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asScaledDecimal: scale 
	"Answer a scaled decimal number, with a fractional precision of scale, 
	approximating the receiver."
	#Numeric.
	"add 200/01/19 For number protocol."
	^ ScaledDecimal newFromNumber: self scale: scale! !

!Number methodsFor: 'double dispatch' stamp: 'RAH 4/25/2000 19:49'!
divideIntoDuration: aDuration 
	"Private - Answer a <Duration>, the result of dividing the receiver into 
	the known Duration, aDuration."
	#DtAndTm.
	"Date and Time Protocols support."
	^ (Smalltalk at: #Duration)
		seconds: aDuration asSeconds / self! !


!Float methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
reciprocal
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	self = 0 ifTrue: ["<- Chg"
		^ (ZeroDivide dividend: self) signal"<- Chg"].
	"<- Chg"
	^ 1.0 / self! !

!Float methodsFor: 'mathematical functions' stamp: 'RAH 4/25/2000 19:49'!
sqrt
	"Answer the square root of the receiver. 
	 Optional. See Object documentation whatIsAPrimitive."
	| exp guess eps delta |
	<primitive: 55>
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	"Newton-Raphson"
	self <= 0.0
		ifTrue: [self = 0.0
				ifTrue: [^ 0.0]
				ifFalse: ["v Chg"
					^ FloatingPointException signal: 'undefined if less than zero.']].
	"first guess is half the exponent"
	exp := self exponent // 2.
	guess := self timesTwoPower: 0 - exp.
	"get eps value"
	eps := guess * Epsilon.
	eps := eps * eps.
	delta := self - (guess * guess) / (guess * 2.0).
	[delta * delta > eps]
		whileTrue: 
			[guess := guess + delta.
			delta := self - (guess * guess) / (guess * 2.0)].
	^ guess! !

!Float methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
hash
	"Hash is reimplemented because = is implemented. Both words of the float 
	are used; 8 bits are removed from each end to clear most of the 
	exponent regardless of the byte ordering. (The bitAnd:'s ensure that the 
	intermediate results do not become a large integer.) Slower than the 
	original version in the ratios 12:5 to 2:1 depending on values. (DNS, 11 
	May, 1997)"
	| tmpInt |
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support 
	receiver = comparand	=> 
	receiver hash = comparand hash"
	tmpInt := self truncated.
	"	<- Chg"
	self = tmpInt ifTrue: ["	<- Chg"
		^ tmpInt hash"	<- Chg"].
	"	<- Chg"
	^ ((self basicAt: 1)
		bitAnd: 16776960)
		+ ((self basicAt: 2)
				bitAnd: 16776960) bitShift: -8! !

!Float methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
adaptToScaledDecimal: receiverScaledDecimal andSend: arithmeticOpSelector 
	"Convert receiverScaledDecimal to a Float and do the arithmetic. 
	receiverScaledDecimal arithmeticOpSelector self."
	#Numeric.
	"add 200/01/19 For ScaledDecimal support."
	^ receiverScaledDecimal asFloat perform: arithmeticOpSelector with: self! !

!Float methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asSpecies: number 
	"Convert number to a Float."
	#Numeric.
	"add 200/01/19 For ANSI <number>support."
	^ number asFloat! !

!Float methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printPaddedWith: aCharacter to: aNumber 
	"Answer the string containing the ASCII representation of the receiver 
	padded on the left with aCharacter to be at least on aNumber 
	integerPart characters and padded the right with aCharacter to be at 
	least anInteger fractionPart characters."
	| aStream digits fPadding fLen iPadding iLen curLen periodIndex |
	#Numeric.
	"2000/03/04  Harmon R. Added Date and Time support"
	aStream := WriteStream on: (String new: 10).
	self printOn: aStream.
	digits := aStream contents.
	periodIndex := digits indexOf: $..
	curLen := periodIndex - 1.
	iLen := aNumber integerPart.
	curLen < iLen
		ifTrue: [iPadding := (String new: (iLen - curLen) asInteger) atAllPut: aCharacter;
					 yourself]
		ifFalse: [iPadding := ''].
	curLen := digits size - periodIndex.
	fLen := (aNumber fractionPart * (aNumber asFloat exponent * 10)) asInteger.
	curLen < fLen
		ifTrue: [fPadding := (String new: fLen - curLen) atAllPut: aCharacter;
					 yourself]
		ifFalse: [fPadding := ''].
	^ iPadding , digits , fPadding! !

!Float methodsFor: 'private' stamp: 'RAH 4/1/2000 08:32'!
asApproximateFraction
	"Answer a Fraction approximating the receiver.  This conversion uses the continued fraction method to approximate a floating point number."
	| num1 denom1 num2 denom2 int frac newD temp |
		#Numeric. "Changed 200/01/19 For <number> support."
	"num1 := self asInteger.		<- Chg"
	num1 := self truncated.	"The first of two alternating numerators."
	denom1 := 1.		"The first of two alternating denominators."
	num2 := 1.		"The second numerator."
	denom2 := 0.		"The second denominator--will update."
	int := num1.		"The integer part of self."
	frac := self fractionPart.	"The fractional part of self."
	[frac = 0] whileFalse: [	"Repeat while the fractional part is not zero."
		newD := 1.0 / frac.	"Take reciprocal of the fractional part."
		int := newD truncated.		"Get the integer part of this and"
		frac := newD fractionPart.	"save fractional part for next time."
		temp := num2.				"Get old numerator and save it"
		num2 := num1.				"Set second numerator to first"
		num1 := num1 * int + temp.	"Update first numerator"
		temp := denom2.				"Get old denominator and save it"
		denom2 := denom1.			"Set second denominator to first"
		denom1 := int * denom1 + temp.		"Update first denominator"
"1e14 ->"		10000000000.0 < denom1 ifTrue: [	"Is ratio past float precision?"
			"Pick which of the two ratios to use"
			num2 = 0.0 ifTrue: [		"Is second denominator 0?"
				^Fraction
					numerator: num1
					denominator: denom1
			].
			^Fraction
				numerator: num2
				denominator: denom2
	]	].
	"If fractional part is zero, return the first ratio"
	denom1 = 1 ifTrue: [	"Am I really an Integer?"
		^num1 "Yes, return Integer result"
	] ifFalse: [			"Otherwise return Fraction result"
		^Fraction
			numerator: num1
			denominator: denom1
	]
! !


!Fraction methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
/ aNumber 
	"Answer the result of dividing the receiver by aNumber."
	#Numeric.
	"Changed 200/01/19 For ANSI support."
	aNumber = 0 ifTrue: [^ (ZeroDivide dividend: self) signal"<-Chg"].
	aNumber isFraction ifTrue: [^ self * aNumber reciprocal].
	^ aNumber adaptToFraction: self andSend: #/! !

!Fraction methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
reciprocal
	"Refer to the comment in Number|reciprocal."
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support."
	numerator = 0 ifTrue: [^ (ZeroDivide dividend: self) signal"<- Chg"].
	numerator = 1 ifTrue: [^ denominator].
	numerator = -1 ifTrue: [^ denominator negated].
	^ Fraction numerator: denominator denominator: numerator! !

!Fraction methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
hash
	"Hash is reimplemented because = is implemented."
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support 
	receiver = comparand	=> 
	receiver hash = comparand hash"
	^ self asFloat hash"	<- Chg"
	"^numerator hash bitXor: denominator hash		<- Prv"! !

!Fraction methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
adaptToScaledDecimal: receiverScaledDecimal andSend: arithmeticOpSelector 
	"Convert receiverScaledDecimal to a Fraction and do the arithmetic. 
	receiverScaledDecimal arithmeticOpSelector self."
	#Numeric.
	"add 200/01/19 For ScaledDecimal support."
	^ receiverScaledDecimal asFraction perform: arithmeticOpSelector with: self! !

!Fraction methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asSpecies: number 
	"Convert number to a Fraction."
	#Numeric.
	"add 200/01/19 For ANSI <number>support."
	^ number asFraction! !

!Fraction methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printOn: aStream 
	#Numeric.
	"Changed 200/01/19 For ANSI <Fraction> removed ( & )"
	numerator printOn: aStream.
	aStream nextPut: $/.
	denominator printOn: aStream! !


!Integer methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
// aNumber 
	| q |
	#Numeric.
	"Changed 200/01/19 For ANSI support."
	aNumber = 0 ifTrue: [^ (ZeroDivide dividend: self) signal"<- Chg"].
	self = 0 ifTrue: [^ 0].
	q := self quo: aNumber.
	"Refer to the comment in Number|//."
	(q negative
		ifTrue: [q * aNumber ~= self]
		ifFalse: [q = 0 and: [self negative ~= aNumber negative]])
		ifTrue: [^ q - 1"Truncate towards minus infinity."]
		ifFalse: [^ q]! !

!Integer methodsFor: 'mathematical functions' stamp: 'RAH 4/25/2000 19:49'!
lcm: operand 
	"Answer the smallest non-negative integer which is evenly divided by 
	both the receiver and operand.  Answer 0 if the receiver and operand 
	are zero."
	#Numeric.
	"Changed 200/03/04 For ANSI <integer> protocol"
	(self = 0 and: [operand = 0])
		ifTrue: ["<-CHG"
			^ 0"<-CHG"].
	"<-CHG"
	^ (self // (self gcd: operand) * operand) abs"<-CHG"! !

!Integer methodsFor: 'bit manipulation' stamp: 'RAH 4/25/2000 19:49'!
bitAt: indexInteger 
	"Answer the value of the bit at indexInteger in the binary 
	representation of the receiver. 
	The result is undefined if the receiver is negative."
	#Numeric.
	"Added 200/03/04 For ANSI <integer> protocol"
	indexInteger <= 0 ifTrue: [^ ArithmeticError signal: 'index less than one'].
	^ ((self bitShift: 1 - indexInteger)
		bitAnd: 1) abs! !

!Integer methodsFor: 'bit manipulation' stamp: 'RAH 4/25/2000 19:49'!
bitAt: indexInteger put: valueLowOrderBitUsedInteger 
	"Set the value of the bit at indexInteger in the binary representation of 
	the receiver.  Tthe value of the bit at position indexInteger is equal to 
	the low order bit of valueLowOrderBitUsedInteger. 
	The result is undefined if either the receiver or 
	valueLowOrderBitUsedInteger is a negative integer."
	| mask |
	#Numeric.
	"Added 200/03/04 For ANSI <integer> protocol"
	indexInteger <= 0 ifTrue: [^ ArithmeticError signal: 'Index less than one.'].
	mask := 1 bitShift: indexInteger - 1.
	(valueLowOrderBitUsedInteger bitAt: 1)
		= 0
		ifTrue: 
			[mask := mask bitXor: self.
			^ self bitAnd: mask].
	^ self bitOr: mask! !

!Integer methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
adaptToScaledDecimal: receiverScaledDecimal andSend: arithmeticOpSelector 
	"Convert me to a ScaledDecimal and do the arithmetic. 
	receiverScaledDecimal arithmeticOpSelector self."
	#Numeric.
	"add 200/01/19 For ScaledDecimal support."
	^ receiverScaledDecimal perform: arithmeticOpSelector with: (self asScaledDecimal: 0)! !

!Integer methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asScaledDecimal: scaleNotUsed 
	"The number of significant digits of the answer is the same as the 
	number of decimal digits in the receiver.  The scale of the answer is 0."
	#Numeric.
	"add 200/01/19 For <integer> protocol."
	^ ScaledDecimal newFromNumber: self scale: 0! !

!Integer methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asSpecies: number 
	"Convert number to a Integer."
	#Numeric.
	"add 200/01/19 For ANSI <number>support."
	^ number asInteger! !

!Integer methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printOn: outputStream base: baseInteger showRadix: flagBoolean 
	"Write a sequence of characters that describes the receiver in radix 
	baseInteger with optional radix specifier. 
	The result is undefined if baseInteger less than 2 or greater than 36."
	| tempString startPos |
	#Numeric.
	"2000/03/04  Harmon R. Added ANSI <integer> protocol"
	tempString := self printStringRadix: baseInteger.
	flagBoolean ifTrue: [^ outputStream nextPutAll: tempString].
	startPos := (tempString indexOf: $r ifAbsent: [self error: 'radix indicator not found.'])
				+ 1.
	self negative ifTrue: [outputStream nextPut: $-].
	outputStream nextPutAll: (tempString copyFrom: startPos to: tempString size)! !

!Integer methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printPaddedWith: aCharacter to: anInteger 
	"Answer the string containing the ASCII representation of the receiver 
	padded on the left with aCharacter to be at least anInteger characters."
	#Numeric.
	"2000/03/04  Harmon R. Added Date and Time support"
	^ self
		printPaddedWith: aCharacter
		to: anInteger
		base: 10! !

!Integer methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printPaddedWith: aCharacter to: anInteger base: aRadix 
	"Answer the string containing the ASCII representation of the receiver 
	padded on the left with aCharacter to be at least anInteger characters."
	| aStream padding digits |
	#Numeric.
	"2000/03/04  Harmon R. Added Date and Time support"
	aStream := WriteStream on: (String new: 10).
	self
		printOn: aStream
		base: aRadix
		showRadix: false.
	digits := aStream contents.
	padding := anInteger - digits size.
	padding > 0 ifFalse: [^ digits].
	^ ((String new: padding) atAllPut: aCharacter;
	 yourself) , digits! !

!Integer methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printStringRadix: baseInteger 
	"Return a string containing a sequence of characters that represents the 
	numeric value of the receiver in the radix specified by the argument.  
	If the receiver is negative, a minus sign ('-') is prepended to the 
	sequence of characters. 
	The result is undefined if baseInteger less than 2 or greater than 36."
	| tempString |
	#Numeric.
	"2000/03/04  Harmon R. Added ANSI <integer> protocol"
	baseInteger = 10
		ifTrue: 
			[tempString := self printStringBase: baseInteger.
			self negative
				ifTrue: [^ '-10r' , (tempString copyFrom: 2 to: tempString size)]
				ifFalse: [^ '10r' , tempString]].
	^ self printStringBase: baseInteger! !

!Integer methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
denominator
	"Answer the denominator of the receiver."
	#Numeric.
	"Changed 200/03/04 For ANSI <rational> support."
	^ 1! !

!Integer methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
numerator
	"Answer the numerator of the receiver."
	#Numeric
	"Changed 200/03/04 For ANSI <rational> support."! !


!Number class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
readRemainderOf: integerPart from: aStream base: base withSign: sign 
	"Read optional fractional part and exponent, and return the final result"
	| value fraction fractionDigits fracpos peekChar scale |
	#Numeric.
	"Changed 200/01/19 For ANSI Numeric Literals support."
	value := integerPart.
	fractionDigits := 0.
	(aStream peekFor: $.)
		ifTrue: ["<integer>.<fraction>"
			(aStream atEnd not and: [aStream peek digitValue between: 0 and: base - 1])
				ifTrue: 
					[fracpos := aStream position.
					fraction := Integer readFrom: aStream base: base.
					fraction := fraction asFloat / (base raisedTo: aStream position - fracpos).
					fractionDigits := aStream position - fracpos.
					value := value asFloat + fraction]
				ifFalse: 
					["oops - just <integer>."
					aStream skip: -1.
					"un-gobble the period"
					^ value * sign"Number readFrom: '3r-22.2'"]].
	peekChar := aStream peek.
	"<number>(e|d|q)<exponent>>"
	peekChar = $e | (peekChar = $d) | (peekChar = $q)
		ifTrue: 
			[aStream next.
			value := value * (base raisedTo: (Integer readFrom: aStream))]
		ifFalse: [peekChar = $s
				ifTrue: 
					["<number>s[<scale>]"
					aStream next.
					(aStream atEnd not and: ["<number>s<scale>"
						aStream peek digitValue between: 0 and: 10])
						ifTrue: [scale := Integer readFrom: aStream].
					scale isNil
						ifTrue: ["<number>s"
							fractionDigits = 0
								ifTrue: ["<integer>s"
									scale := 0]
								ifFalse: ["<integer>.<fraction>s"
									scale := fractionDigits]].
					value := ScaledDecimal newFromNumber: value scale: scale]].
	(value isFloat
		and: [value = 0.0 and: [sign = -1]])
		ifTrue: [^ Float negativeZero]
		ifFalse: [^ value * sign]! !


!Float class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
one
	#Numeric.
	"add 200/01/19 For <number> protocol support."
	^ 1.0! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
denormalized
	"Answer a Boolean indication of whether the characterized floating point 
	object representation allows denormalized values."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	^ true! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
emax
	"Answer an Integer, the largest exponent of the characterized floating 
	point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	^ 1024"11 bit exponent."! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
emin
	"Answer an Integer, the smallest exponent of the characterized floating 
	point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	^ -1023! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
epsilon
	"Answer a Float, the maximum relative spacing in the characterized 
	floating point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	^ (self radix raisedTo: 1 - self precision) asFloat! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
fmax
	"Answer a Float, the largest value allowed by the characterized floating 
	point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	"Same value as MaxVal."
	^ (1 - (Float radix raisedTo: Float precision negated) * (Float radix raisedTo: Float emax)) asFloat! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
fmin
	"Answer a Float, the minimum value allowed by the characterized 
	floating point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	self denormalized ifTrue: [^ self fminDenormalized].
	^ self fminNormalized! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
fminDenormalized
	"Answer a Float, the minimum denormalized value allowed by the 
	characterized floating point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	#todo.
	"Correct return value, but calculation differs from that specified in ANSI 
	standard.  I think asFloat has minor error. 
	From ''Numerical Computation Guide: 2 - IEEE Arithmetic'' 
	min positive subnormal number	00000000 00000001 4.9406564584124654e-324 
	 From 5.6.8.8 Message:	fminDenormalized 
	self radix raisedTo: (self emin - self precision)"
	^ (self radix raisedTo: self emin - self precision + 2) asFloat! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
fminNormalized
	"Answer a Float, the minimum normalized value allowed by the 
	characterized floating point object representation."
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	#todo.
	"ANSI standard Definition: <floatCharacterization> 
	. . . This satisfies the ISO/IEC 10967 floating point characterization 
	requirement fminN, and is equal to 
	 
	self radix raisedTo: (self emin - 1). 
	 
	(Float radix raisedTo: (Float emin - 1)) asFloat -> 5.56268464626801e-309 
	 
	From ''Numerical Computation Guide: 2 - IEEE Arithmetic'' 
	min positive normal number	: 00100000 00000000	2.2250738585072014e-308 
	 
	(Float radix raisedTo: (Float emin + 1)) asFloat -> 2.2250738585072e-308 
	 ^ changed 
	agrees with Guide.  I will use plus as it appears to be correct."
	^ (self radix raisedTo: self emin + 1) asFloat! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
precision
	"Answer an Integer, the precision of the characterized floating point 
	object representation."
	#Numeric.
	"add 200/01/19 For <Float> protocol."
	#todo.
	"Verify correct - I am unsure of it."
	^ 53"52 bit mantissa (normalized)  plus an assumed high order 1 bit"! !

!Float class methodsFor: 'characterization' stamp: 'RAH 4/25/2000 19:49'!
radix
	"Answer an Integer, the radix of the characterized floating point object 
	representation. "
	#Numeric.
	"Added 2000/03/04 For <floatCharacterization> protocol."
	^ 2! !


!Fraction class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
one
	#Numeric.
	"add 200/01/19 For <number> protocol support."
	^ self numerator: 1 denominator: 1! !


!Integer class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
one
	#Numeric.
	"add 200/01/19 For <number> protocol support."
	^ 1! !


!PositionableStream methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
zisEmpty
	"Returns a Boolean indicating whether there are any sequence values in 
	the receiver."
	#Stream.
	"Changed 2000/04/08 For ANSI <sequencedStream> protocol."
	#todo.
	"Breaks String keywords."
	^ readLimit = 0! !


!Preferences class methodsFor: 'standard preferences' stamp: 'RAH 4/25/2000 19:49'!
autoDeclineMembership
	#Fundmntl.
	"Added 2000/04/08 For ANSI support."
	^ self valueOfFlag: #autoDeclineMembership! !


!ReadStream methodsFor: 'file open/close' stamp: 'RAH 4/25/2000 19:49'!
close
	"Disassociate a stream from its backing store."
	#Stream.
	"Added 2000/04/08 For ANSI <sequencedStream> protocol."
	collection := nil.
	readLimit := nil.
	position := nil! !


!RuleDate methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
applyRuleTo: daysNamedInMonthList 
	"Private - Answer the day of the month selected from dayOfMonth list 
	by applying the receiver's rule."
	^ daysNamedInMonthList perform: selectionRule! !

!RuleDate methodsFor: 'private' stamp: 'BG 2/22/2003 22:28'!
basicUpdateForMonth: mm year: yyyy 
	"Private - Answer the receiver after updating by applying the rule for 
	the month, mm, of year, yyyy."
	| dayByRule daysNamedInMonth firstDayNamed day |
	firstDayNamed := self
				dayInMonth: mm
				year: yyyy
				ofFirstDayNamed: dayOfWeek.
	day := (Date
				newDay: firstDayNamed
				month: mm
				year: yyyy) day.
	
     self julianDayNumber:
		 (Date  newDay: day
				year: yyyy) julianDayNumber.
	daysNamedInMonth := (firstDayNamed to: self daysInMonth by: 7) asArray.
	dayByRule := self applyRuleTo: daysNamedInMonth.
	self julianDayNumber:
		 (Date
				newDay: dayByRule
				month: mm
				year: yyyy) julianDayNumber! !

!RuleDate methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
dayInMonth: monthIn year: yearIn ofFirstDayNamed: dayNameIn 
	"Private - Answer, the day in the month, monthIn, of year, yearIn, of 
	the first day named, dayNameIn."
	| frstDayNdx dayName firstDay |
	dayName := dayNameIn asSymbol.
	frstDayNdx := (Date firstWeekdayOfMonth: monthIn year: yearIn)
				- 1.
	frstDayNdx = 0 ifTrue: [frstDayNdx := frstDayNdx + 7].
	firstDay := 1.
	(Date nameOfDay: frstDayNdx)
		= dayName
		ifFalse: 
			[firstDay := 1 + (Date dayOfWeek: dayName) - frstDayNdx.
			firstDay < 1 ifTrue: [firstDay := firstDay + 7]].
	^ firstDay! !

!RuleDate methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
setDayOfWeek: dayName selectionRule: positionName 
	"Private - Set dayOfWeek to dayName, and selectionRule to positionName. 
	 
	Parameters 
	dayName		<Symbol>		captured 
	positionName	<Symbol>		captured
	"
	dayOfWeek := dayName.
	selectionRule := positionName! !

!RuleDate methodsFor: 'updating' stamp: 'RAH 4/25/2000 19:49'!
updateForYear: yyyy 
	"Update the receiver by applying the rule for its month number of 
	year, yyyy. 
	 
	Definition: <RuleDate> 
	Parameters 
	yyyy	<Integer>	captured
	"
	self basicUpdateForMonth: self monthIndex year: yyyy! !

!RuleDate methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isRuleDate
	#DtAndTm.
	"Added 2000/04/08 To use DNU mod."
	^ true! !


!RuleDate class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
first: dayName inMonth: monthIn year: yearIn 
	"Answer a date that is the first day of the week (Sunday, etc), dayName, 
	in month, monthIn, of year, yearIn.  Example: 
	 
	RuleDate first: 'Monday' inMonth: 'April' year: 2000 
	 
	Note: The month may be an index or a month name. 
	The year may be specified as the actual number of years since the 
	beginning of the Roman calendar or the  number of years since 1900, 
	or a two digit date from 1900.  1/1/01 will NOT mean 2001. 
	Definition: <RuleDate factory> 
	Parameters 
	dayName	<readableString>				captured 
	monthIn	<readableString> | <Integer>	captured 
	yearIn		<Integer>					captured 
	Return Values 
	<RuleDate>		new 
	Errors 
	Day name is not a String and a valid day of the week 
	Month is not an Integer 1 - 12. or a valid month name String
	"
	| mmInt newRuleDate |
	mmInt := self getValidMonthNumber: monthIn.
	newRuleDate := self newDayOfWeek: dayName selectionRule: 'first'.
	newRuleDate basicUpdateForMonth: mmInt year: yearIn.
	^ newRuleDate! !

!RuleDate class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
last: dayName inMonth: monthIn year: yearIn 
	"Answer a date that is the last day of the week (Sunday, etc), dayName, 
	in month, monthIn, of year, yearIn.  Example: 
	 
	RuleDate last: 'Monday' inMonth: 'April' year: 2000 
	 
	Note: The month may be an index or a month name. 
	The year may be specified as the actual number of years since the 
	beginning of the Roman calendar or the  number of years since 1900, 
	or a two digit date from 1900.  1/1/01 will NOT mean 2001. 
	Definition: <RuleDate factory> 
	Parameters 
	dayName	<readableString>				captured 
	monthIn	<readableString> | <Integer>	captured 
	yearIn		<Integer>					captured 
	Return Values 
	<RuleDate>		new 
	Errors 
	Day name is not a String and a valid day of the week 
	Month is not an Integer 1 - 12. or a valid month name String
	"
	| mmInt newRuleDate |
	mmInt := self getValidMonthNumber: monthIn.
	newRuleDate := self newDayOfWeek: dayName selectionRule: 'last'.
	newRuleDate basicUpdateForMonth: mmInt year: yearIn.
	^ newRuleDate! !

!RuleDate class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
getValidMonthNumber: monthIn 
	"Private - Answer the month number of monthIn if it is a month name 
	String, else monthIn as the month number  if it is an Integer, else 
	signal an error. 
	 
	Definition: <RuleDate factory> 
	Parameters 
	monthIn	<readableString> | <Integer>	captured 
	Return Values 
	<RuleDate>		new 
	Errors 
	Month is not an Integer 1 - 12. or a valid month name String
	"
	monthIn isInteger
		ifTrue: 
			[(monthIn between: 1 and: 12)
				ifTrue: [^ monthIn].
			^ Error signal: 'Month must be 1 - 12.'].
	(monthIn isMemberOf: String)
		ifTrue: [^ self indexOfMonth: monthIn].
	^ Error signal: 'Month must be an Integer 1 - 12 or a month name String.'! !

!RuleDate class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
getValidSelectionRule: positionName 
	"Private - Answer the selection position (first, last) in the list of day of 
	the week,  Report an error if positionName is not one of (first, last)."
	| positionSymbol |
	(positionName isMemberOf: String)
		ifFalse: [^ Error signal: 'Position name: "' , positionName , '" is not a String.'].
	positionSymbol := positionName asLowercase asSymbol.
	(#(first last ) includes: positionSymbol)
		ifFalse: [^ Error signal: 'Position name: "' , positionName , '" is not valid.'].
	^ positionSymbol! !

!RuleDate class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
newDayOfWeek: dayName selectionRule: positionName 
	"Private - Answer an uncreated rule date with the dayOfWeek (Sunday, 
	etc) set to dayName, and selectionRule (first, last) set to positionName. 
	Note: Must be updated to create the date. 
	 
	Parameters 
	dayName		<readableString>		captured 
	positionName	<readableString>		captured 
	Return Values 
	<RuleDate>		new
	"
	| daySymbol newRuleDate positionSymbol |
	(dayName isMemberOf: String)
		ifFalse: [^ Error signal: 'Day name: "' , dayName , '" is not a String.'].
	daySymbol := dayName asLowercase.
	daySymbol at: 1 put: (daySymbol at: 1) asUppercase.
	daySymbol := daySymbol asSymbol.
	(WeekDayNames includes: daySymbol)
		ifFalse: [^ Error signal: 'Day name: "' , dayName , '" is not valid.'].
	positionSymbol := self getValidSelectionRule: positionName.
	newRuleDate := super new.
	newRuleDate setDayOfWeek: daySymbol selectionRule: positionSymbol.
	^ newRuleDate! !


!RuleIndexDate methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
applyRuleTo: daysNamedInMonthList 
	"Private - Answer the day of the month selected from dayOfMonth list 
	by applying the receiver's rule."
	^ daysNamedInMonthList at: selectionRule! !


!RuleIndexDate class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
getValidSelectionRule: dayListIndex 
	"Private - Answer the dayListIndex position in the list of all days named 
	(Sunday, etc) in a month,  Report an error if dayListIndex does not 
	represent an <Integer>."
	dayListIndex isInteger ifFalse: [^ self error: 'Not an Integer.'].
	^ dayListIndex! !

!RuleIndexDate class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
indexed: dayName inMonth: monthIn year: yearIn at: dayListIndex 
	"Answer a date that is at the dayListIndex position of the list of day of 
	the week (Sunday, etc), dayName, in month, monthIn, of year, yearIn.  
	Example:  
	RuleIndexDate indexed: 'Sunday' inMonth: 'April' year: 2000 at: 2. 
	 
	Note: The month may be an index or a month name. 
	The year may be specified as the actual number of years since the 
	beginning of the Roman calendar or the  number of years since 1900, 
	or a two digit date from 1900.  1/1/01 will NOT mean 2001. 
	 
	Definition: <RuleDate factory> 
	Parameters 
	dayName		<readableString>				captured 
	monthIn		<readableString> | <Integer>	captured 
	yearIn			<Integer>					captured 
	dayListIndex		<Integer>					captured 
	Return Values 
	<RuleDate>		new 
	Errors 
	Day name is not a String and a valid day of the week 
	Month is not an Integer 1 - 12. or a valid month name String
	"
	| mmInt newRuleDate |
	mmInt := self getValidMonthNumber: monthIn.
	newRuleDate := self newDayOfWeek: dayName selectionRule: dayListIndex.
	newRuleDate basicUpdateForMonth: mmInt year: yearIn.
	^ newRuleDate! !


!RuleSelectionCodeDate methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
applyRuleTo: daysNamedInMonthList 
	"Private - Answer the day of the month selected from dayOfMonth list 
	by applying the receiver's rule."
	^ selectionRule value: daysNamedInMonthList! !


!RuleSelectionCodeDate class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
getValidSelectionRule: selectionBlock 
	"Private - Answer the selectionBlock to select the date given the 
	dayName (Sunday, etc) dayOfMonth list,  Report an error if 
	selectionBlock does not represent a <monadicValuable> block."
	((selectionBlock isMemberOf: BlockContext)
		and: [selectionBlock argumentCount = 1])
		ifFalse: [^ self error: 'Not an <monadicValuable> block.'].
	^ selectionBlock! !

!RuleSelectionCodeDate class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
selected: dayName inMonth: monthIn year: yearIn byCode: selectionBlock 
	"Answer a date selected by selectionBlock given the dayName (Sunday, 
	etc) dayOfMonth list, in month, monthIn, of year, yearIn.  Example: 
	 
	Standard Time starts on Sunday, 
	so the Daylight Time end is the previous day: 
	RuleSelectionCodeDate 
	selected: 'Sunday' 
	inMonth: 'October' 
	year: 2000 
	byCode: [ :sundaysList | 
	(sundaysList last) - 1 
	]. 
	 
	Note: The selected date need not be in the dayOfMonth list, but may be 
	relative to a named day. 
	The month may be an index or a month name. 
	The year may be specified as the actual number of years since the 
	beginning of the Roman calendar or the  number of years since 1900, 
	or a two digit date from 1900.  1/1/01 will NOT mean 2001. 
	Definition: <RuleDate factory> 
	Parameters 
	dayName		<readableString>				captured 
	monthIn		<readableString> | <Integer>	captured 
	yearIn			<Integer>					captured 
	selectionBlock	<monadicValuable>			captured 
	Return Values 
	<RuleDate>		new 
	Errors 
	Day name is not a String and a valid day of the week 
	Month is not an Integer 1 - 12. or a valid month name String
	"
	| mmInt newRuleDate |
	mmInt := self getValidMonthNumber: monthIn.
	newRuleDate := self newDayOfWeek: dayName selectionRule: selectionBlock.
	newRuleDate basicUpdateForMonth: mmInt year: yearIn.
	^ newRuleDate! !


!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
* operand 
	"Implementation of Number 'arithmetic' method."
	operand isScaledDecimal ifTrue: [^ ScaledDecimal newFromNumber: fraction * operand asFraction scale: (scale max: operand scale)].
	^ operand adaptToScaledDecimal: self andSend: #*! !

!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
+ operand 
	"Implementation of Number 'arithmetic' method."
	operand isScaledDecimal ifTrue: [^ ScaledDecimal newFromNumber: fraction + operand asFraction scale: (scale max: operand scale)].
	^ operand adaptToScaledDecimal: self andSend: #+! !

!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
- operand 
	"Implementation of Number 'arithmetic' method."
	operand isScaledDecimal ifTrue: [^ ScaledDecimal newFromNumber: fraction - operand asFraction scale: (scale max: operand scale)].
	^ operand adaptToScaledDecimal: self andSend: #-! !

!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
/ operand 
	"Implementation of Number 'arithmetic' method."
	#ScalDec.
	"Protocol: ANSI <number>."
	operand = 0 ifTrue: [^ (ZeroDivide dividend: self) signal].
	operand isScaledDecimal ifTrue: [^ ScaledDecimal newFromNumber: fraction / operand asFraction scale: (scale max: operand scale)].
	^ operand adaptToScaledDecimal: self andSend: #/! !

!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
// operand 
	"Answer the integer quotient after dividing the receiver by operand 
	with truncation towards negative infinity."
	^ fraction // operand! !

!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
negated
	"Reimplementation of Number 'arithmetic' method."
	^ ScaledDecimal newFromNumber: fraction negated scale: scale! !

!ScaledDecimal methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
reciprocal
	"Reimplementation of Number 'arithmetic' method."
	self = 0 ifTrue: [^ (ZeroDivide dividend: 1) signal].
	^ ScaledDecimal newFromNumber: fraction reciprocal scale: scale! !

!ScaledDecimal methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
< operand 
	"Implementation of Number 'comparing' method."
	operand isScaledDecimal ifTrue: [^ fraction < operand asFraction].
	^ operand adaptToScaledDecimal: self andSend: #<! !

!ScaledDecimal methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
= comparand 
	"Implementation of Number 'comparing' method."
	comparand isNumber ifFalse: [^ false].
	comparand isScaledDecimal ifTrue: [^ fraction = comparand asFraction].
	^ comparand adaptToScaledDecimal: self andSend: #=! !

!ScaledDecimal methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
hash
	"Reimplementation of Object 'comparing' method."
	^ fraction hash! !

!ScaledDecimal methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
isZero
	"Answer whether the receiver is equal to its class' zero"
	^ fraction numerator = 0! !

!ScaledDecimal methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
adaptToFraction: receiver andSend: arithmeticOpSelector 
	"Convert me to a Fraction and do the arithmetic. 
	receiver arithmeticOpSelector self."
	^ receiver perform: arithmeticOpSelector with: fraction! !

!ScaledDecimal methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
adaptToInteger: receiver andSend: arithmeticOpSelector 
	"Convert receiver to a ScaledDecimal and do the arithmetic. 
	receiver arithmeticOpSelector self."
	^ (receiver asScaledDecimal: 0)
		perform: arithmeticOpSelector with: self! !

!ScaledDecimal methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asFloat
	"Reimplementation - Number 'converting' method."
	^ fraction asFloat! !

!ScaledDecimal methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asFraction
	"Implementation - Number 'converting' method."
	^ fraction! !

!ScaledDecimal methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asScaledDecimal: scaleIn 
	"Reimplementation - Number 'converting' method."
	^ ScaledDecimal newFromNumber: fraction scale: scaleIn! !

!ScaledDecimal methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
asSpecies: number 
	"Convert number to a ScaledDecimal."
	#Numeric.
	"add 200/01/19 For ANSI <number>support."
	^ ScaledDecimal newFromNumber: number scale: scale! !

!ScaledDecimal methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
denominator
	"Private - Answer an Integer, the denominator part of the receiver."
	^ fraction denominator! !

!ScaledDecimal methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
numerator
	"Private - Answer an Integer, the numerator part of the receiver."
	^ fraction numerator! !

!ScaledDecimal methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
scale
	"Private - Answer a integer which represents the total number of digits 
	used to represent the fraction part of the receiver, including trailing 
	zeroes. "
	^ scale! !

!ScaledDecimal methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
setFraction: fractionIn scale: scaleIn 
	"Private - Set the fraction to fractionIn and the total number of digits 
	used to represent the fraction part of the receiver, including trailing 
	zeroes, to the Integer scaleIn."
	fraction := fractionIn.
	scale := scaleIn! !

!ScaledDecimal methodsFor: 'truncation and round off' stamp: 'RAH 4/25/2000 19:49'!
fractionPart
	"Answer the fractional part of the receiver."
	^ ScaledDecimal newFromNumber: fraction fractionPart scale: scale! !

!ScaledDecimal methodsFor: 'truncation and round off' stamp: 'RAH 4/25/2000 19:49'!
integerPart
	"Answer the fractional part of the receiver."
	^ ScaledDecimal newFromNumber: fraction integerPart scale: scale! !

!ScaledDecimal methodsFor: 'truncation and round off' stamp: 'RAH 4/25/2000 19:49'!
truncated
	"Reimplementation of Number 'truncation and round off' method."
	^ fraction truncated! !

!ScaledDecimal methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isScaledDecimal
	"Reimplementation - Number 'testing' method."
	^ true! !

!ScaledDecimal methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printOn: aStream 
	"Reimplementation - Object 'printing' method."
	| aFraction tmpFractionPart |
	self < 0 ifTrue: [aStream nextPut: $-].
	aFraction := fraction abs.
	aStream nextPutAll: aFraction truncated printString.
	scale = 0 ifTrue: [^ aStream nextPutAll: 's0'].
	aStream nextPut: $..
	tmpFractionPart := aFraction fractionPart.
	1 to: scale
		do: 
			[:dummy | 
			tmpFractionPart := tmpFractionPart * 10.
			aStream nextPut: (Character digitValue: tmpFractionPart truncated).
			tmpFractionPart := tmpFractionPart fractionPart].
	aStream nextPut: $s.
	scale printOn: aStream! !

!ScaledDecimal methodsFor: 'printing' stamp: 'RAH 4/25/2000 19:49'!
printString
	"Reimplementation - Number 'printing' method."
	| tempStream |
	tempStream := WriteStream on: (String new: 10).
	self printOn: tempStream.
	^ tempStream contents! !

!ScaledDecimal methodsFor: 'mathematical functions' stamp: 'RAH 4/25/2000 19:49'!
squared
	"Reimplementation - Number 'mathematical functions' method."
	"not used ->"
	^ ScaledDecimal newFromNumber: fraction squared scale: scale! !


!ScaledDecimal class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
newFromNumber: aNumber scale: scaleIn 
	"Answer a new instance of me."
	| temp |
	temp := self basicNew.
	temp setFraction: aNumber asFraction scale: scaleIn.
	^ temp! !

!ScaledDecimal class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
one
	"Answer the receiver's representation of one."
	^ self newFromNumber: 1 scale: 0! !

!ScaledDecimal class methodsFor: 'constants' stamp: 'RAH 4/25/2000 19:49'!
zero
	"Answer the receiver's representation of zero."
	^ self newFromNumber: 0 scale: 0! !


!SequenceableCollection methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
after: target 
	"Answer the object immediately following the first element which is 
	equivalent to target in the receiver."
	| index |
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	index := self indexOf: target ifAbsent: [self error: 'no element in the receiver is equivalent to target.'].
	index := index + 1.
	index > self size ifTrue: [self error: 'element equal to target is the last element in the receiver.'].
	^ self at: index! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
at: index 
	"Answer the element at the position index in the receiver."
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	index isInteger ifFalse: [^ self errorNonIntegerIndex].
	(index between: 1 and: self size)
		ifFalse: [^ self errorSubscriptBounds: index].
	^ super at: index! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
before: target 
	"Answer the object immediately preceding the first element which is 
	equivalent to target in the receiver."
	| index |
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	index := self indexOf: target ifAbsent: [self error: 'no element in the receiver is equivalent to target.'].
	index = 1 ifTrue: [self error: 'element equal to target is the first element in the receiver.'].
	^ self at: index - 1! !

!SequenceableCollection methodsFor: 'comparing' stamp: 'RAH 4/25/2000 19:49'!
= otherCollection 
	"Answer true if the receiver is equivalent to the <otherCollection>. 
	    First test for identity, then rule out different species and sizes of 
	collections.     As a last resort, examine each element of the receiver 
	and the <otherCollection>."
	| size |
	#Collectn.
	"Changed 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	self == otherCollection ifTrue: [^ true].
	self class == otherCollection class ifFalse: [^ false].
	"<- Chg."
	(size := self size) = otherCollection size ifFalse: [^ false].
	1 to: size
		do: [:index | (self at: index)
				= (otherCollection at: index) ifFalse: [^ false]].
	^ true! !

!SequenceableCollection methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
reverse
	"Answer a collection with the elements of the receiver arranged in 
	reverse order."
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	^ self reversed! !

!SequenceableCollection methodsFor: 'copying' stamp: 'RAH 4/25/2000 19:49'!
copyReplaceFrom: start to: stop withObject: replacement 
	"Answer a new collection conforming to the same protocols as the 
	receiver, in which the elements of the receiver between start and stop 
	inclusive have been replaced with replacement. 
	Note: This message can be used to insert, append, or replace.  There are 
	three cases: 
	1. If stop = start - 1 , and start is less than or equal to the size of the 
	receiver, then replacement is inserted between the elements at index 
	stop and start.  None of the receiver's elements are replaced. 
	 
	'12345' copyReplaceFrom: 3 to: 2 withObject: $x -> '12x345' 
	 
	2. If stop = the size of the receiver and start = stop + 1, then the 
	operation is an append, and replacement is placed at the end of the 
	new collection. 
	 
	'12345' copyReplaceFrom: 6 to: 5 withObject: $x -> '12345x' 
	 
	3. Otherwise, the operation is a replacement, and each of the receiver's 
	elements in the given range is replaced by replacement. 
	 
	'12345' copyReplaceFrom: 2 to: 4 withObject: $x -> '1xxx5' 
	 
	The parameters start and stop must be non-negative."
	| tempSeqCollection |
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	(self class meetsElementTypeConstraints: replacement)
		ifFalse: [^ self class errorElementType].
	start > (self size + 1) ifTrue: [^ self error: 'start must be <= to receiver''s size + 1.'].
	start < 1 ifTrue: [^ self error: 'start must be >= 1.'].
	stop > self size ifTrue: [^ self error: 'stop must be <= to receiver''s size.'].
	stop < (start - 1) ifTrue: [^ self error: 'stop must be >= start - 1.'].
	(stop = (start - 1) and: [start <= self size])
		ifTrue: 
			["1. Insert."
			tempSeqCollection := self species new: self size + 1.
			1 to: stop do: [:ndx | tempSeqCollection at: ndx put: (self at: ndx)].
			tempSeqCollection at: start put: replacement.
			start to: self size do: [:ndx | tempSeqCollection at: ndx + 1 put: (self at: ndx)].
			^ tempSeqCollection].
	(stop = self size and: [start = (stop + 1)])
		ifTrue: 
			["2. Append."
			tempSeqCollection := self species new: start.
			1 to: stop do: [:ndx | tempSeqCollection at: ndx put: (self at: ndx)].
			tempSeqCollection at: start put: replacement.
			^ tempSeqCollection].
	"3. Replacement."
	tempSeqCollection := self species new: self size.
	1 to: start - 1 do: [:ndx | tempSeqCollection at: ndx put: (self at: ndx)].
	start to: stop do: [:ndx | tempSeqCollection at: ndx put: replacement].
	stop + 1 to: self size do: [:ndx | tempSeqCollection at: ndx put: (self at: ndx)].
	^ tempSeqCollection! !

!SequenceableCollection methodsFor: 'copying' stamp: 'RAH 4/25/2000 19:49'!
copyReplacing: target withObject: replacement 
	"Answer a new collection conforming to the same protocols as the 
	receiver, in which any occurrences of target are replaced by 
	replacement. "
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	(self class meetsElementTypeConstraints: replacement)
		ifFalse: [^ self class errorElementType].
	^ self
		collect: [:element | element = target
				ifTrue: [replacement]
				ifFalse: [element]]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'RAH 4/25/2000 19:49'!
from: start to: stop do: operation 
	"For those elements of the receiver between positions start and stop, 
	inclusive, evaluate operation with each element of the receiver."
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	self checkFrom: start to: stop.
	start to: stop do: [:index | operation value: (self at: index)]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'RAH 4/25/2000 19:49'!
from: start to: stop keysAndValuesDo: operation 
	"For those elements of the receiver between positions start and stop, 
	inclusive, evaluate operation with an element of the receiver as the 
	first argument and the element's position (index) as the second."
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	self checkFrom: start to: stop.
	start to: stop do: [:index | operation value: (self at: index)
			value: index]! !

!SequenceableCollection methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
checkFrom: start to: stop 
	"Private - Check that the arguments start and stop are positive Integers, 
	are valid indices for the receiver collection.  If they are invalid, report 
	the error."
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> support."
	start isInteger & start positive & stop isInteger & stop positive ifFalse: [^ self error: 'start and stop must be positive integers.'].
	(start between: 1 and: self size)
		ifFalse: [^ self errorOutOfBounds start].
	(stop between: 1 and: self size)
		ifFalse: [^ self errorOutOfBounds stop]! !


!ArrayedCollection methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
replaceFrom: start to: stop withObject: anObject 
	"Replace the elements of the receiver between start and stop inclusive 
	with replacementElement.  Answer the receiver."
	| index |
	#Collectn.
	"Changed 2000/04/08 For ANSI <sequencedCollection> protocol."
	index := start.
	[index <= stop]
		whileTrue: 
			[self at: index put: anObject.
			index := index + 1].
	^ anObject! !


!ByteArray methodsFor: 'enumerating' stamp: 'RAH 4/25/2000 19:49'!
collect: transformer 
	"Re-implementation of Collection <enumerating> #collect:."
	| classObj newElement result |
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> protocol."
	classObj := self class.
	result := self species new: self size.
	1 to: self size
		do: 
			[:index | 
			newElement := transformer value: (self at: index).
			(classObj meetsElementTypeConstraints: newElement)
				ifFalse: [^ classObj errorElementType].
			result at: index put: newElement].
	^ result! !


!ByteArray class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
errorElementType
	"Private - At least one element did not meet the type constraints of the 
	receiver collection."
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	self error: 'All elements must be 0 <= Integer <= 255.'! !

!ByteArray class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
meetsElementTypeConstraints: elementObject 
	"Private - Answer a Boolean, true if elementObject meets the element 
	type constraints of the receiver collection, false otherwise."
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	elementObject isInteger
		ifTrue: [(elementObject between: 0 and: 255)
				ifFalse: [^ false]]
		ifFalse: [^ false].
	^ true! !

!ByteArray class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
with: element1 
	"Create a collection initially containing the argument element."
	#Collectn.
	"Changed 2000/04/08 For ANSI <ByteArray factory> protocol."
	(self meetsElementTypeConstraints: element1)
		ifFalse: [^ self errorElementType].
	^ super with: element1! !

!ByteArray class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
with: element1 with: element2 
	"Create a collection initially containing the argument elements."
	#Collectn.
	"Changed 2000/04/08 For ANSI <ByteArray factory> protocol."
	(self meetsElementTypeConstraints: element1)
		& (self meetsElementTypeConstraints: element2) ifFalse: [^ self errorElementType].
	^ super with: element1 with: element2! !

!ByteArray class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
with: element1 with: element2 with: element3 
	"Create a collection initially containing the argument elements."
	#Collectn.
	"Changed 2000/04/08 For ANSI <ByteArray factory> protocol."
	(self meetsElementTypeConstraints: element1)
		& (self meetsElementTypeConstraints: element2) & (self meetsElementTypeConstraints: element3) ifFalse: [^ self errorElementType].
	^ super
		with: element1
		with: element2
		with: element3! !

!ByteArray class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
with: element1 with: element2 with: element3 with: element4 
	"Create a collection initially containing the argument elements."
	#Collectn.
	"Changed 2000/04/08 For ANSI <ByteArray factory> protocol."
	(self meetsElementTypeConstraints: element1)
		& (self meetsElementTypeConstraints: element2) & (self meetsElementTypeConstraints: element3) & (self meetsElementTypeConstraints: element4) ifFalse: [^ self errorElementType].
	^ super
		with: element1
		with: element2
		with: element3
		with: element4! !

!ByteArray class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
withAll: newElements 
	"Create a collection containing all the elements of newElements."
	#Collectn.
	"Changed 2000/04/08 For ANSI <ByteArray factory> protocol."
	(newElements allSatisfy: [:element | self meetsElementTypeConstraints: element])
		ifFalse: [^ self errorElementType].
	^ super withAll: newElements! !


!OrderedCollection methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
replaceFrom: start to: stop withObject: anObject 
	"Replace the elements of the receiver between start and stop inclusive 
	with replacementElement.  Answer the receiver."
	| index |
	#Collectn.
	"Changed 2000/04/08 For ANSI <sequencedCollection> protocol."
	index := start.
	[index <= stop]
		whileTrue: 
			[self at: index put: anObject.
			index := index + 1].
	^ anObject! !

!OrderedCollection methodsFor: 'adding' stamp: 'RAH 4/25/2000 19:49'!
add: newElement beforeIndex: index 
	"Add newElement to the receiver immediately before the element at 
	position index."
	#Collectn.
	"Changed 2000/04/08 For ANSI <OrderedCollection> protocol."
	self checkIndex: index.
	self insert: newElement before: index.
	^ newElement! !

!OrderedCollection methodsFor: 'adding' stamp: 'RAH 4/25/2000 19:49'!
addAll: newElements after: target 
	"Add each element of newElements to the receiver immediately after the 
	first element in the receiver which is equivalent to target.  Answer 
	newElements. "
	| index |
	#Collectn.
	"Changed 2000/04/08 For ANSI <OrderedCollection> protocol."
	index := self indexOf: target.
	index = 0 ifTrue: [^ self errorNotFound].
	^ self addAll: newElements afterIndex: index! !

!OrderedCollection methodsFor: 'adding' stamp: 'RAH 4/25/2000 19:49'!
addAll: newElements afterIndex: index 
	"Insert the elements of newElements in the receiver immediately after 
	the element at position index.  Answer newElements."
	| newCollection newIndex newSize oldSize |
	#Collectn.
	"Changed 2000/04/08 For ANSI <OrderedCollection> protocol."
	index = 0 ifTrue: ["If index is equal to 0, 
		newElements are inserted at the beginning of the receiver."
		^ self addAllFirst: newElements].
	self checkIndex: index.
	oldSize := self size.
	newSize := oldSize + newElements size.
	newCollection := Array new: newSize.
	newIndex := 1.
	1 to: index
		do: 
			[:oldIndex | 
			newCollection at: newIndex put: (self at: oldIndex).
			newIndex := newIndex + 1].
	newElements
		do: 
			[:newElement | 
			newCollection at: newIndex put: newElement.
			newIndex := newIndex + 1].
	index + 1 to: oldSize
		do: 
			[:oldIndex | 
			newCollection at: newIndex put: (self at: oldIndex).
			newIndex := newIndex + 1].
	array := newCollection.
	firstIndex := 1.
	lastIndex := newSize.
	^ newElements! !

!OrderedCollection methodsFor: 'adding' stamp: 'RAH 4/25/2000 19:49'!
addAll: newElements before: target 
	"Add each element of newElements to the receiver immediately before the 
	first element in the receiver which is equivalent to target.  Answer 
	newElements. "
	| index |
	#Collectn.
	"Changed 2000/04/08 For ANSI <OrderedCollection> protocol."
	index := self indexOf: target.
	index = 0 ifTrue: [^ self errorNotFound].
	index = 1 ifTrue: [^ self addAllFirst: newElements].
	^ self addAll: newElements afterIndex: index - 1! !

!OrderedCollection methodsFor: 'adding' stamp: 'RAH 4/25/2000 19:49'!
addAll: newElements beforeIndex: index 
	"Insert the elements of newElements in the receiver immediately before 
	the element at position index.  Answer newElements."
	#Collectn.
	"Changed 2000/04/08 For ANSI <OrderedCollection> protocol."
	index = 1 ifTrue: [^ self addAllFirst: newElements].
	^ self addAll: newElements afterIndex: index - 1! !

!OrderedCollection methodsFor: 'removing' stamp: 'RAH 4/25/2000 19:49'!
removeAtIndex: index 
	"Remove the element of the receiver at position index, and answer the 
	removed element."
	| element |
	#Collectn.
	"Changed 2000/04/08 For ANSI <sequencedContractibleCollection> protocol."
	self checkIndex: index.
	element := self at: index.
	self removeIndex: index.
	^ element! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'RAH 4/25/2000 19:49'!
do: aBlock 
	"Override the superclass for performance reasons."
	| index |
	firstIndex isNil ifTrue: [^ self].
	index := firstIndex.
	[index <= lastIndex]
		whileTrue: 
			[aBlock value: (array at: index).
			index := index + 1]! !

!OrderedCollection methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
checkIndex: ndx 
	"Private - Check that ndx is a positive integer less than or equal to the 
	receiver's size.  Report the error if not."
	#Collectn.
	"Changed 2000/04/08 For ANSI <sequencedContractibleCollection> support."
	ndx isInteger ifFalse: [^ self error: 'Must be indexed by integers'].
	(ndx between: 1 and: self size)
		ifFalse: [^ self error: 'Index not 1 <= ' , ndx printString , ' <= size']! !


!RunArray class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
scanFrom: strm 
	"Read the style section of a fileOut or sources file.  nextChunk has 
	already been done.  We need to return a RunArray of TextAttributes of 
	various kinds."
	| rr vv aa this |
	#Numeric.
	"Changed 200/01/19 to use Integer class>>#readFrom: instead of Number 
	(fontNumber confused it)."
	(strm peekFor: $()
		ifFalse: [^ nil].
	rr := OrderedCollection new.
	
	[strm skipSeparators.
	strm peekFor: $)]
		whileFalse: [rr add: (Integer readFrom: strm)].
	"<- Chg."
	vv := OrderedCollection new.
	"Value array"
	aa := OrderedCollection new.
	"Attributes list"
	[(this := strm next) == nil]
		whileFalse: 
			[this == $,
				ifTrue: 
					[vv add: aa asArray.
					aa := OrderedCollection new].
			this == $f ifTrue: [aa add: (TextFontChange new fontNumber: (Integer readFrom: strm))].
			"<- Chg."
			this == $b ifTrue: [aa add: TextEmphasis bold].
			this == $i ifTrue: [aa add: TextEmphasis italic].
			this == $u ifTrue: [aa add: TextEmphasis underlined].
			this == $= ifTrue: [aa add: TextEmphasis struckOut].
			this == $n ifTrue: [aa add: TextEmphasis normal].
			this == $- ifTrue: [aa add: (TextKern kern: -1)].
			this == $+ ifTrue: [aa add: (TextKern kern: 1)].
			this == $c ifTrue: [aa add: (TextColor scanFrom: strm)].
			"color"
			this == $L ifTrue: [aa add: (TextLink scanFrom: strm)].
			"L not look like 1"
			this == $R ifTrue: [aa add: (TextURL scanFrom: strm)].
			"R capitalized so it can follow a number"
			this == $q ifTrue: [aa add: (TextSqkPageLink scanFrom: strm)].
			this == $d ifTrue: [aa add: (TextDoIt scanFrom: strm)]].
	aa size > 0 ifTrue: [vv add: aa asArray].
	^ self runs: rr asArray values: vv asArray" 
	RunArray scanFrom: (ReadStream on: '(14 50 312)f1,f1b,f1LInteger +;i')
	"! !


!SmallInteger methodsFor: 'arithmetic' stamp: 'RAH 4/25/2000 19:49'!
quo: aNumber 
	"Primitive. Divide the receiver by the argument and answer with the 
	result.  Round the result down towards zero to make it a whole integer. 
	Fail if the argument is 0 or is not a SmallInteger. Optional.  See Object 
	documentation whatIsAPrimitive."
	<primitive: 13>
	#Numeric.
	"Changed 200/01/19 For ANSI <number> support"
	aNumber = 0 ifTrue: [^ (ZeroDivide dividend: self) signal"	<- Chg"
		"^self error: 'Attempt to divide by zero'"].
	(aNumber isMemberOf: SmallInteger)
		ifTrue: [self primitiveFailed]
		ifFalse: [^ super quo: aNumber]! !


!SortedCollection methodsFor: 'accessing' stamp: 'BG 2/22/2003 22:20'!
sortBlock
	"Answer the blockContext which is the criterion for sorting elements of the receiver."
	#Collectn.
	"Changed 2000/06/23 Harmon, R. For ANSI <SortedCollection> protocol."

	sortBlock isNil ifFalse: [^ sortBlock].
	^ self class defaultSortBlock! !


!String methodsFor: 'accessing' stamp: 'RAH 4/25/2000 19:49'!
indexOfSubCollection: sub 
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	^ self
		indexOfSubCollection: sub
		startingAt: 1
		ifAbsent: [0]! !

!String methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
subStrings
	"Answer an array of the substrings that compose the receiver."
	#Collectn.
	"Added 2000/04/08 For ANSI <readableString> protocol."
	^ self substrings! !

!String methodsFor: 'converting' stamp: 'RAH 4/25/2000 19:49'!
subStrings: separators 
	"Answer an array containing the substrings in the receiver separated 
	by the elements of separators."
	| char result sourceStream subString |
	#Collectn.
	"Changed 2000/04/08 For ANSI <readableString> protocol."
	(separators allSatisfy: [:element | element isKindOf: Character])
		ifFalse: [^ self error: 'separators must be Characters.'].
	sourceStream := ReadStream on: self.
	result := OrderedCollection new.
	subString := String new.
	[sourceStream atEnd]
		whileFalse: 
			[char := sourceStream next.
			(separators includes: char)
				ifTrue: [subString notEmpty
						ifTrue: 
							[result add: subString copy.
							subString := String new]]
				ifFalse: [subString := subString , (String with: char)]].
	subString notEmpty ifTrue: [result add: subString copy].
	^ result asArray! !

!String methodsFor: 'testing' stamp: 'RAH 4/25/2000 19:49'!
isString
	#DtAndTm.
	"Added 2000/04/08 To use DNU mod."
	^ true! !


!String class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
errorElementType
	"Private - At least one element did not meet the type constraints of the 
	receiver collection."
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	self error: 'All elements must be Character.'! !

!String class methodsFor: 'private' stamp: 'RAH 4/25/2000 19:49'!
meetsElementTypeConstraints: element 
	"Private - Answer a Boolean, true if element meets the element type 
	constraints of the receiver collection, false otherwise."
	#Collectn.
	"Added 2000/04/08 For ANSI <collection> support."
	element isCharacter ifFalse: [^ false].
	^ true! !


!WriteStream methodsFor: 'file open/close' stamp: 'RAH 4/25/2000 19:49'!
close
	"Disassociate a stream from its backing store."
	#Stream.
	"Added 2000/04/08 For ANSI <sequencedStream> protocol."
	self flush.
	collection := nil.
	readLimit := nil.
	position := nil! !

!WriteStream methodsFor: 'file open/close' stamp: 'RAH 4/25/2000 19:49'!
flush
	"Update a stream's backing store."
	#Stream! !


!FileStream class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
read: fileId 
	"Returns a read file stream that reads text from the file with the given 
	name. 
	Implementation Note: temp. version for SIFMod."
	#SIFMod.
	"Protocol: <FileStream factory>."
	^ self concreteStream readOnlyFileNamed: (self fullName: fileId)"Chg SQ."! !

!FileStream class methodsFor: 'instance creation' stamp: 'RAH 4/25/2000 19:49'!
write: fileId 
	"Returns a write file stream that writes text to the file with the given 
	name. 
	Implementation Note: temp. version for SIFMod."
	#SIFMod.
	"Protocol: <FileStream factory>."
	^ self newFileNamed: fileId! !

ReadWriteStream removeSelector: #close!

!WriteStream reorganize!
('accessing' contents next nextPut: nextPutAll: size)
('character writing' cr crtab crtab: ensureASpace ensureNoSpace nextPutKeyword:withArg: peekLast space tab tab:)
('file open/close' close flush)
('fileIn/Out' nextChunkPut: nextChunkPutWithStyle: timeStamp)
('positioning' position: reset resetToStart setToEnd)
('printing' store:)
('private' braceArray braceArray: on: on:from:to: pastEndPut: with: withAttribute:do: withAttributes:do:)
!


!ByteArray reorganize!
('accessing' atAllPut: byteAt: byteAt:put: byteSize bytesPerElement)
('platform independent access' longAt:bigEndian: longAt:put:bigEndian: shortAt:bigEndian: shortAt:put:bigEndian: unsignedLongAt:bigEndian: unsignedLongAt:put:bigEndian: unsignedShortAt:bigEndian: unsignedShortAt:put:bigEndian:)
('converting' asByteArray asString)
('private' asByteArrayPointer asExternalPointer defaultElement replaceFrom:to:with:startingAt:)
('comparing' hash)
('enumerating' collect:)
('external access' booleanAt: booleanAt:put: doubleAt: doubleAt:put: floatAt: floatAt:put: integerAt:put:size:signed: integerAt:size:signed: isExternalAddress pointerAt: pointerAt:put: signedByteAt: signedByteAt:put: signedCharAt: signedCharAt:put: signedLongAt: signedLongAt:put: signedLongLongAt: signedLongLongAt:put: signedShortAt: signedShortAt:put: structAt:length: structAt:put:length: unsignedByteAt: unsignedByteAt:put: unsignedCharAt: unsignedCharAt:put: unsignedLongAt: unsignedLongAt:put: unsignedLongLongAt: unsignedLongLongAt:put: unsignedShortAt: unsignedShortAt:put: voidAt: voidAt:put:)
!


!ReadStream reorganize!
('accessing' ascii binary next next: next:into:startingAt: nextPut: size upTo: upToEnd)
('file open/close' close)
('private' on:from:to:)
!

Fraction removeSelector: #raisedToInteger:!
Float removeSelector: #raisedTo:!
Duration initialize!
DateAndTime initialize!
